"""Generate Java code for de/serialization based on the intermediate representation."""

import io
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    indent_but_first_line,
    Error,
    Stripped,
    Identifier,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


def _generate_enum_to_and_from_string(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the methods for de/serializing enumeration from/to a string."""
    blocks = []  # type: List[Stripped]

    name = java_naming.enum_name(enumeration.name)

    # region To-string-map

    enum_to_string_blocks = []  # type: List[Stripped]

    for literal in enumeration.literals:
        literal_name = java_naming.enum_literal_name(literal.name)

        literal_string = java_common.string_literal(literal.value)

        enum_to_string_blocks.append(
            Stripped(f"""temp.put({name}.{literal_name}, {literal_string});""")
        )

    enum_to_string_mapping = "\n".join(enum_to_string_blocks)

    to_str_map_name = java_naming.property_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    to_str_map = Stripped(
        f"""\
private static final Map<{name}, String> {to_str_map_name};
static {{
{I}final Map<{name}, String> temp = new HashMap<>();

{I}{indent_but_first_line(enum_to_string_mapping, I)}

{I}if (!temp.keySet().containsAll(Arrays.asList({name}.values()))) {{
{II}throw new IllegalStateException("Unmapped {name}");
{I}}}

{I}{to_str_map_name} = Collections.unmodifiableMap(temp);
}}"""
    )

    blocks.append(to_str_map)

    # endregion

    # region To-string-method

    to_str_name = java_naming.method_name(Identifier("to_string"))

    to_str_writer = io.StringIO()
    to_str_writer.write(
        f"""\
/**
 * Retrieve the string representation of {{@code that}}.
 *
 * <p>If {{@code that}} is not a valid literal, return {{@code Optional#empty()}}.
 */
public static Optional<String> {to_str_name}({name} that)
{{
{I}return Optional.ofNullable(that).map({to_str_map_name}::get);
}}"""
    )

    blocks.append(Stripped(to_str_writer.getvalue()))

    # endregion

    # region From-string-map

    string_to_enum_blocks = []  # type: List[Stripped]

    for literal in enumeration.literals:
        literal_name = java_naming.enum_literal_name(literal.name)

        literal_string = java_common.string_literal(literal.value)

        string_to_enum_blocks.append(
            Stripped(f"""temp.put({literal_string}, {name}.{literal_name});""")
        )

    string_to_enum_mapping = "\n".join(string_to_enum_blocks)

    from_str_map_name = java_naming.private_property_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    from_str_map = Stripped(
        f"""\
private static final Map<String, {name}> {from_str_map_name};
static {{
{I}final Map<String, {name}> temp = new HashMap<>();

{I}{indent_but_first_line(string_to_enum_mapping, I)}

{I}if (!temp.values().containsAll(Arrays.asList({name}.values()))) {{
{II}throw new IllegalStateException("Unmapped {name}");
{I}}}

{I}{from_str_map_name} = Collections.unmodifiableMap(temp);
}}"""
    )

    blocks.append(from_str_map)

    # endregion

    # region From-string-method

    from_str_name = java_naming.method_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    from_str_writer = io.StringIO()
    from_str_writer.write(
        f"""\
/**
 * Parse the string representation of {{@link {name}}}.
 *
 * <p>If {{@code text}} is not a valid string representation
 * of a literal of {{@link {name}}},
 * return {{@code Optional#empty()}}.
 */
public static Optional<{name}> {from_str_name}(String text)
{{
{I}{name} value = {from_str_name}.get(text);
{I}if (value == null) {{
{II}return Optional.empty();
{I}}} else {{
{II}return Optional.of(value);
{I}}}
}}"""
    )

    blocks.append(Stripped(from_str_writer.getvalue()))

    # endregion

    return Stripped("\n\n".join(blocks))


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Java code for the general serialization.

    The ``package`` defines the root Java package.
    """
    stringification_blocks = []  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        stringification_blocks.append(
            _generate_enum_to_and_from_string(enumeration=enum)
        )

    stringification_code = Stripped("\n\n".join(stringification_blocks))

    code = Stripped(
        f"""\
{java_common.WARNING}

package {package}.stringification;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import {package}.types.enums.*;

public class Stringification {{
{I}{indent_but_first_line(stringification_code, I)}
}}

{java_common.WARNING}"""
    )

    return f"{code}\n", None



"""Generate C# code for enhancing model classes."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_delegate_method(method: intermediate.Method) -> Stripped:
    """Generate the delegated method to ``_instance``."""
    returns = (
        csharp_common.generate_type(method.returns, our_type_qualifier=Stripped("Aas"))
        if method.returns is not None
        else "void"
    )

    arg_types_names = [
        (
            csharp_common.generate_type(
                arg.type_annotation, our_type_qualifier=Stripped("Aas")
            ),
            csharp_naming.argument_name(arg.name),
        )
        for arg in method.arguments
    ]

    method_name = csharp_naming.method_name(method.name)

    return_prefix = "return " if method.returns is not None else ""

    if len(method.arguments) == 0:
        return Stripped(
            f"""\
public {returns} {method_name}()
{{
{I}{return_prefix}_instance.{method_name}();
}}"""
        )

    arguments_definition = ",\n".join(
        f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
    )

    arguments_delegation = ",\n".join(arg_name for _, arg_name in arg_types_names)

    return Stripped(
        f"""\
public {returns} {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
)
{{
{I}{return_prefix}_instance.{method_name}(
{II}{indent_but_first_line(arguments_delegation, II)}
{I});
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@require(lambda cls: not cls.is_implementation_specific)
def _generate_enhanced_class(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the structure for the enhanced concrete class."""
    enhanced_name = csharp_naming.class_name(Identifier(f"enhanced_{cls.name}"))
    interface_name = csharp_naming.interface_name(cls.name)

    blocks = [
        Stripped(f"private readonly Aas.{interface_name} _instance;"),
        Stripped(
            f"""\
public {enhanced_name}(
{I}Aas.{interface_name} instance,
{I}TEnhancement enhancement
) : base(enhancement)
{{
{I}_instance = instance;
}}"""
        ),
    ]  # type: List[Stripped]

    for prop in cls.properties:
        prop_type = csharp_common.generate_type(prop.type_annotation)
        prop_name = csharp_naming.property_name(prop.name)

        blocks.append(
            Stripped(
                f"""\
public {prop_type} {prop_name}
{{
{I}get => _instance.{prop_name};
{I}set => _instance.{prop_name} = value;
}}"""
            )
        )

    # region OverXOrEmpty getter

    for prop in cls.properties:
        if isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        ) and isinstance(prop.type_annotation.value, intermediate.ListTypeAnnotation):
            prop_name = csharp_naming.property_name(prop.name)
            items_type = csharp_common.generate_type(
                prop.type_annotation.value.items, our_type_qualifier=Stripped("Aas")
            )

            blocks.append(
                Stripped(
                    f"""\
public IEnumerable<{items_type}> Over{prop_name}OrEmpty()
{{
{I}return _instance.Over{prop_name}OrEmpty();
}}"""
                )
            )

    # endregion

    for method in cls.methods:
        blocks.append(_generate_delegate_method(method))

    visit_name = csharp_naming.method_name(Identifier(f"visit_{cls.name}"))

    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    blocks.extend(
        [
            Stripped(
                f"""\
public IEnumerable<Aas.IClass> DescendOnce()
{{
{I}return _instance.DescendOnce();
}}"""
            ),
            Stripped(
                f"""\
public IEnumerable<Aas.IClass> Descend()
{{
{I}return _instance.Descend();
}}"""
            ),
            Stripped(
                f"""\
public void Accept(Aas.Visitation.IVisitor visitor)
{{
{I}visitor.{visit_name}(_instance);
}}"""
            ),
            Stripped(
                f"""\
public void Accept<TContext>(
{I}Visitation.IVisitorWithContext<TContext> visitor,
{I}TContext context
)
{{
{I}visitor.{visit_name}(_instance, context);
}}"""
            ),
            Stripped(
                f"""\
public T Transform<T>(Visitation.ITransformer<T> transformer)
{{
{I}return transformer.{transform_name}(_instance);
}}"""
            ),
            Stripped(
                f"""\
public T Transform<TContext, T>(
{I}Visitation.ITransformerWithContext<TContext, T> transformer,
{I}TContext context
)
{{
{I}return transformer.{transform_name}(_instance, context);
}}"""
            ),
        ]
    )

    writer = io.StringIO()
    writer.write(
        f"""\
public class {enhanced_name}<TEnhancement>
{I}: Enhanced<TEnhancement>, Aas.{interface_name}
{I}where TEnhancement : class
{{
"""
    )

    if len(blocks) > 0:
        for i, block in enumerate(blocks):
            if i > 0:
                writer.write("\n\n")
            writer.write(textwrap.indent(block, I))
    else:
        writer.write("// No properties to wire to _instance.")

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


@require(lambda cls: not cls.is_implementation_specific)
def _generate_transform(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the transform method to wrap the instance with an enhancement."""
    blocks = [
        Stripped(
            f"""\
if (that is Enhanced<TEnhancement>)
{{
{I}throw new System.ArgumentException(
{II}$"The instance has been already enhanced: {{that}}"
{I});
}}"""
        )
    ]  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)
        prop_name = csharp_naming.property_name(prop.name)

        wrap_stmt: Stripped

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            # We can not enhance primitive types; nothing to do here.
            continue

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                # We can not enhance enumerations; nothing to do here.
                continue

            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                # We can not enhance primitive types; nothing to do here.
                continue

            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                value_interface_name = csharp_naming.interface_name(
                    type_anno.our_type.name
                )
                transformed_name = csharp_naming.variable_name(
                    Identifier(f"transformed_{prop.name}")
                )
                casted_name = csharp_naming.variable_name(
                    Identifier(f"casted_{prop.name}")
                )
                wrap_stmt = Stripped(
                    f"""\
var {transformed_name} = Transform(
{I}that.{prop_name}
);
var {casted_name} = (
{I}{transformed_name} as Aas.{value_interface_name}
) ?? throw new System.InvalidOperationException(
{I}"Expected the transformed value to be a {value_interface_name}, " +
{I}$"but got: {{{transformed_name}}}"
);
that.{prop_name} = {casted_name};"""
                )
            else:
                assert_never(type_anno.our_type)
        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            # fmt: off
            assert (
                isinstance(type_anno.items, intermediate.OurTypeAnnotation)
                and isinstance(
                    type_anno.items.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass)
               )
            ), (
                "(mristin, 2023-02-08) We handle only lists of classes in"
                "the enhancing at the moment. The meta-model does not contain "
                "any other lists, so we wanted to keep the code as simple as "
                "possible, and avoid unrolling. Please contact the developers "
                "if you need this feature."
            )
            # fmt: on

            item_interface_name = csharp_naming.interface_name(
                type_anno.items.our_type.name
            )

            wrap_stmt = Stripped(
                f"""\
that.{prop_name} = (
{I}that.{prop_name}
{I}.Select(
{II}(item) => {{
{III}var transformed = Transform(item);
{III}return (
{IIII}transformed as Aas.{item_interface_name}
{III}) ?? throw new System.InvalidOperationException(
{IIII}"Expected the transformed item to be a {item_interface_name}, " +
{IIII}$"but got: {{transformed}}"
{III});
{II}}}
{I})
).ToList();"""
            )
        else:
            assert_never(type_anno.our_type)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            wrap_stmt = Stripped(
                f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(wrap_stmt, I)}
}}"""
            )

        blocks.append(wrap_stmt)

    enhanced_name = csharp_naming.class_name(Identifier(f"enhanced_{cls.name}"))

    blocks.append(
        Stripped(
            f"""\
var enhancement = _enhancementFactory(that);
return (enhancement == null)
{I}? that
{I}: new {enhanced_name}<TEnhancement>(
{II}that,
{II}enhancement
{I});"""
        )
    )

    interface_name = csharp_naming.interface_name(cls.name)
    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    blocks_joined = "\n\n".join(blocks)

    return Stripped(
        f"""\
public override Aas.IClass {transform_name}(
{I}Aas.{interface_name} that
)
{{
{I}{indent_but_first_line(blocks_joined, I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_wrapper(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transformer that wraps an instance with the enhancement."""
    errors = []  # type: List[Error]
    blocks = [
        Stripped(
            "private readonly System.Func<Aas.IClass, TEnhancement?> "
            "_enhancementFactory;"
        ),
        Stripped(
            f"""\
internal Wrapper(
{I}System.Func<Aas.IClass, TEnhancement?> enhancementFactory
)
{{
{I}_enhancementFactory = enhancementFactory;
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Enhancing/Wrap/{cls.name}.cs"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            blocks.append(code)
            continue

        blocks.append(_generate_transform(cls=cls))

    writer = io.StringIO()
    writer.write(
        f"""\
internal class Wrapper<TEnhancement>
{I}: Aas.Visitation.AbstractTransformer<Aas.IClass>
{I}where TEnhancement : class
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code for enhancing model classes with custom wraps.

    The ``namespace`` defines the AAS C# namespace.
    """
    enhancing_blocks = [
        Stripped(
            f"""\
public abstract class Enhanced<TEnhancement> where TEnhancement : class
{{
{I}// ReSharper disable once InconsistentNaming
{I}protected readonly TEnhancement _enhancement;

{I}protected Enhanced(TEnhancement enhancement)
{I}{{
{II}_enhancement = enhancement;
{I}}}

{I}internal TEnhancement _getEnhancement()
{I}{{
{II}return _enhancement;
{I}}}
}}"""
        )
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Enhancing/Enhanced/{cls.name}.cs"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            assert code is not None
            enhancing_blocks.append(code)
        else:
            code, error = _generate_enhanced_class(cls=cls)
            if error is not None:
                errors.append(error)
                continue

            assert code is not None
            enhancing_blocks.append(code)

    wrapper, wrapper_errors = _generate_wrapper(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if wrapper_errors is not None:
        errors.extend(wrapper_errors)

    if len(errors) > 0:
        return None, errors

    assert wrapper is not None
    enhancing_blocks.append(wrapper)

    enhancing_blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Unwrap enhancements from the wrapped instances.
/// </summary>
/// <typeparam name="TEnhancement">type of the enhancement</typeparam>
public class Unwrapper<TEnhancement> where TEnhancement : class
{{
{I}/// <summary>
{I}/// Unwrap the given model instance.
{I}/// </summary>
{I}/// <param name="that">model instance to be unwrapped</param>
{I}/// <returns>
{I}/// Enhancement, or <c>null</c> if <paramref name="that" />
{I}/// has not been wrapped yet.
{I}/// </returns>
{I}public TEnhancement? Unwrap(Aas.IClass that)
{I}{{
{II}// ReSharper disable once SuspiciousTypeConversion.Global
{II}var enhanced = that as Enhanced<TEnhancement>;
{II}return enhanced?._getEnhancement();
{I}}}

{I}/// <summary>
{I}/// Unwrap the given model instance.
{I}/// </summary>
{I}/// <param name="that">model instance to be unwrapped</param>
{I}/// <returns>
{I}/// Enhancement wrapped around <paramref name="that" />
{I}/// </returns>
{I}/// <exception cref="System.ArgumentException">
{I}/// Thrown when <paramref name="that" /> has not been wrapped yet
{I}/// </exception>
{I}public TEnhancement MustUnwrap(Aas.IClass that)
{I}{{
{II}return Unwrap(that) ?? throw new System.ArgumentException(
{III}$"Expected the instance to have been wrapped, but it was not: {{that}}"
{II});
{I}}}
}}  // public class Unwrapper"""
        )
    )

    enhancing_blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Wrap and unwrap the instances of model classes with enhancement.
/// </summary>
/// <typeparam name="TEnhancement">type of the enhancement</typeparam>
public class Enhancer<TEnhancement>
{I}: Unwrapper<TEnhancement>
{I}where TEnhancement : class
{{
{I}private readonly Wrapper<TEnhancement> _wrapper;

{I}/// <param name="enhancementFactory">
{I}/// <para>how to enhance the instances.</para>
{I}///
{I}/// <para>If it returns <c>null</c>, the instance will not be wrapped. However,
{I}/// the wrapping will continue recursively.</para>
{I}///</param>
{I}public Enhancer(
{II}System.Func<Aas.IClass, TEnhancement?> enhancementFactory
{I})
{I}{{
{II}_wrapper = new Wrapper<TEnhancement>(enhancementFactory);
{I}}}

{I}/// <summary>
{I}/// Wrap the instance with an enhancement.
{I}/// </summary>
{I}/// <remarks>
{I}/// Double wraps are not allowed to prevent runtime leakage.
{I}///
{I}/// If you use references to the instance objects, you have to update them
{I}/// after the wrapping, as the wrapping is recursive.
{I}/// </remarks>
{I}/// <param name="that">model instance to be wrapped</param>
{I}/// <returns>
{I}/// <paramref name="that" /> instance wrapped recursively with enhancements
{I}/// </returns>
{I}/// <exception cref="System.ArgumentException">
{I}/// Thrown when <paramref name="that" /> has been already wrapped
{I}/// </exception>
{I}public Aas.IClass Wrap(
{II}Aas.IClass that
{I})
{I}{{
{II}var wrapped = _wrapper.Transform(that);
{II}return (
{III}wrapped
{II}) ?? throw new System.InvalidOperationException(
{III}"Expected the wrapped instance to be an instance of IClass, " +
{III}$"but got: {{wrapped}}"
{II});
{I}}}
}}  // public class Enhancer"""
        )
    )

    using_directives = []  # type: List[Stripped]
    using_directives.extend(
        csharp_common.generate_using_aas_directive_if_necessary(namespace)
    )

    using_directives.append(
        Stripped(
            """\
using System.Collections.Generic;  // can't alias
using System.Linq;  // can't alias"""
        )
    )

    blocks = [
        csharp_common.WARNING,
        Stripped("\n".join(using_directives)),
    ]

    enhancing_writer = io.StringIO()
    enhancing_writer.write(
        f"""\
namespace {namespace}
{{
{I}/// <summary>
{I}/// Allow for enhancing of our model classes with custom wraps.
{I}/// </summary>
{I}public static class Enhancing
{I}{{
"""
    )

    for i, enhancing_block in enumerate(enhancing_blocks):
        if i > 0:
            enhancing_writer.write("\n\n")

        enhancing_writer.write(textwrap.indent(enhancing_block, II))

    enhancing_writer.write(f"\n{I}}}  // public static class Enhancing")
    enhancing_writer.write(f"\n}}  // namespace {namespace}")

    blocks.append(Stripped(enhancing_writer.getvalue()))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None



"""Generate C++ functions to de/serialize an instance from a JSON value."""

from aas_core_codegen.cpp.jsonization import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate C++ code of a virtual machine for matching regular expressions."""

from aas_core_codegen.cpp.revm import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Generate C++ code of a virtual machine for matching regular expressions."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
    INDENT7 as IIIIIII,
    INDENT8 as IIIIIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(library_namespace: Stripped) -> str:
    """Generate the C++ header of a virtual machine for matching regexes."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.REVM_NAMESPACE}")

    include_guard_var = cpp_common.include_guard_var(namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            """\
#pragma warning(push, 0)
#include <cstdint>
#include <memory>
#include <string>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup {cpp_common.REVM_NAMESPACE} Match regular expressions using a multi-threaded virtual machine.
 *
 * The implementation in the standard library has exponential time complexity, so it was
 * a major blocker for most of the practical inputs. For example, see this bug report:
 * https://gcc.gnu.org/bugzilla/show_bug.cgi?id=93502
 *
 * This implementation is based on Ken Thompson's approach which uses a virtual
 * machine to match regular expressions. The original technique has been
 * published in:
 * Thompson, K., "Regular expression search algorithm", ACM 11(6) (June 1968)
 *
 * We followed a very clear and concise blog post which described in detail:
 * https://swtch.com/~rsc/regexp/regexp2.html
 *
 * The ideas for additional instructions were taken from:
 * https://www.codeproject.com/Articles/5256833/Regex-as-a-Tiny-Threaded-Virtual-Machine
 * @{{
 */
namespace {cpp_common.REVM_NAMESPACE} {{

enum class InstructionKind : std::uint8_t {{
{I}Char,
{I}Set,
{I}NotSet,
{I}Any,
{I}Match,
{I}Jump,
{I}Split,
{I}End
}};

/**
 * Represent an instruction of the virtual machine.
 */
struct Instruction {{
{I}// NOTE (mristin):
{I}// We avoid RTTI for performance reasons, and use our own enumerator instead.
{I}virtual InstructionKind kind() const = 0;

{I}virtual ~Instruction() = default;
}};

/**
 * Match a single character.
 *
 * If the character on the String Pointer does not match the `character`, stop this
 * thread as it failed. Otherwise, move the String Pointer to the next character,
 * and the Program Counter to the next instruction.
 */
struct InstructionChar : Instruction {{
{I}wchar_t character;

{I}explicit InstructionChar(wchar_t a_character);
{I}InstructionKind kind() const override;
{I}~InstructionChar() override = default;
}};

std::string to_string(const InstructionChar& instruction);

/**
 * Define a character range.
 */
struct Range {{
{I}wchar_t first;
{I}wchar_t last;

{I}Range(wchar_t a_first, wchar_t a_last);
}};

std::string to_string(const Range& range);

/**
 * Check whether the character is in any of the given character ranges.
 *
 * @return true if the character is in any of the ranges
 */
bool CharacterInRanges(
{I}const std::vector<Range>& ranges,
{I}wchar_t character
);

/**
 * Match a set of characters.
 *
 * If the character on the String Pointer *is not* in the given set, stop this
 * thread as it failed. Otherwise, move the String Pointer to the next character,
 * and the Program Counter to the next instruction.
 */
struct InstructionSet : Instruction {{
{I}std::vector<Range> ranges;

{I}explicit InstructionSet(std::vector<Range> a_ranges);
{I}InstructionKind kind() const override;
{I}~InstructionSet() override = default;
}};

std::string to_string(const InstructionSet& instruction);

/**
 * Match an out-of-set character.
 *
 * If the character on the String Pointer *is* in the given set, stop this
 * thread as it failed. Otherwise, move the String Pointer to the next character,
 * and the Program Counter to the next instruction.
 */
struct InstructionNotSet : Instruction {{
{I}std::vector<Range> ranges;

{I}explicit InstructionNotSet(std::vector<Range> a_ranges);
{I}InstructionKind kind() const override;
{I}~InstructionNotSet() override = default;
}};

std::string to_string(const InstructionNotSet& instruction);

/**
 * Match any character.
 */
struct InstructionAny : Instruction {{
{I}InstructionKind kind() const override;
{I}~InstructionAny() override = default;
}};

std::string to_string(const InstructionAny&);

/**
 * Stop this thread and signal that we found a match.
 */
struct InstructionMatch : Instruction {{
{I}InstructionKind kind() const override;
{I}~InstructionMatch() override = default;
}};

std::string to_string(const InstructionMatch&);

/**
 * Jump to the indicated position in the program.
 */
struct InstructionJump : Instruction {{
{I}size_t target;

{I}explicit InstructionJump(size_t a_target);
{I}InstructionKind kind() const override;
{I}~InstructionJump() override = default;
}};

std::string to_string(const InstructionJump& instruction);

/**
 * Split the program in two threads, both jumping to different locations. The string
 * pointer is kept as-is.
 */
struct InstructionSplit : Instruction {{
{I}size_t first_target;
{I}size_t second_target;

{I}explicit InstructionSplit(size_t a_first_target, size_t a_second_target);
{I}InstructionKind kind() const override;
{I}~InstructionSplit() override = default;
}};

std::string to_string(const InstructionSplit& instruction);

/**
 * Match the end-of-input.
 */
struct InstructionEnd : Instruction {{
{I}InstructionKind kind() const override;
{I}~InstructionEnd() override = default;
}};

std::string to_string(const InstructionEnd&);

std::string to_string(const Instruction& instruction);

std::string to_string(
{I}const std::vector<std::unique_ptr<Instruction> >& instructions
);

/**
 * Try to match the program against the text.
 * @return true if the text matches
 */
bool Match(
{I}const std::vector<std::unique_ptr<Instruction> >& program,
{I}const std::wstring& text
);

}}  // namespace {cpp_common.REVM_NAMESPACE}
/**@}}*/"""
        ),
        cpp_common.generate_namespace_closing(library_namespace),
        cpp_common.WARNING,
        Stripped(f"#endif  // {include_guard_var}"),
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    library_namespace: Stripped,
) -> str:
    """Generate the C++ implementation of a virtual machine for matching regexes."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.REVM_NAMESPACE}")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/revm.hpp"

#pragma warning(push, 0)
#include <algorithm>
#include <iomanip>
#include <sstream>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        Stripped(
            f"""\
/**
 * Represent the character such that it can be printed in the console.
 *
 * To that end, we escape the character if it is out of the printable ASCII set.
 *
 * This function is mostly meant for debugging purposes.
 */
std::string RepresentWCharacter(wchar_t character) {{
{I}switch (character) {{
{II}case L'\\\\':return "\\\\\\\\";
{II}case L'"':return "\\\\\\"";
{II}case L'\\'':return "\\\\'";
{II}case L'\\t':return "\\\\t";
{II}case L'\\n':return "\\\\n";
{II}case L'\\r':return "\\\\r";
{II}default: break;
{I}}}

{I}if (26 <= character && character <= 126) {{
{II}return std::string(1, static_cast<char>(character));
{I}}}

{I}static const char* digits = "0123456789ABCDEF";
{I}size_t digit_count = sizeof(wchar_t) * 2;

{I}std::string result;
{I}result.resize(digit_count + 2);
{I}result[0] = L'\\\\';
{I}result[1] = L'u';

{I}for (size_t i = 0, j = (digit_count - 1) * 4; i < digit_count; ++i, j -= 4) {{
{II}const size_t digit_i = (character >> j) & 0x0f;
{II}result[i + 2] = digits[digit_i];
{I}}}

{I}return result;
}}

/**
 * Represent the wide string for debugging purposes where it is printed to the console.
 *
 * To that end, we escape the individual characters if they are out of printable ASCII
 * set.
 */
std::string RepresentWString(const std::wstring& text) {{
{I}std::vector<std::string> parts;
{I}parts.reserve(text.size());
{I}for (const wchar_t character : text) {{
{II}parts.emplace_back(RepresentWCharacter(character));
{I}}}

{I}size_t size = 0;
{I}for (const std::string& part : parts) {{
{II}size += part.size();
{I}}}

{I}std::string result;
{I}result.reserve(size);
{I}for (const std::string& part : parts) {{
{II}result.append(part);
{I}}}
{I}return result;
}}

InstructionChar::InstructionChar(
{I}wchar_t a_character
) :
{I}character(a_character) {{
{I}// Intentionally empty.
}}

InstructionKind InstructionChar::kind() const {{
{I}return InstructionKind::Char;
}}

std::string to_string(const InstructionChar& instruction) {{
{I}return common::Concat(
{II}"char '",
{II}RepresentWCharacter(instruction.character),
{II}"'"
{I});
}}

Range::Range(
{I}wchar_t a_first,
{I}wchar_t a_last
) :
{I}first(a_first),
{I}last(a_last) {{
{I}// NOTE (mristin):
{I}// We are aware that exceptions in constructors should be avoided to prevent
{I}// bug related to uninitialized object state. However, in this case, we do not
{I}// see any risk for such a mistake.
{I}if (a_first > a_last) {{
{II}throw std::invalid_argument(
{III}common::Concat(
{IIII}"The first character in a character range, ",
{IIII}RepresentWCharacter(a_first),
{IIII}", is larger than the last character in the range, ",
{IIII}RepresentWCharacter(a_last)
{III})
{II});
{I}}}
}}

std::string to_string(const Range& range) {{
{I}if (range.first == range.last) {{
{II}return RepresentWCharacter(range.first);
{I}}}

{I}return common::Concat(
{II}RepresentWCharacter(range.first),
{II}"-",
{II}RepresentWCharacter(range.last)
{I});
}}

InstructionSet::InstructionSet(
{I}std::vector<Range> a_ranges
) :
{I}ranges(std::move(a_ranges)) {{
{I}// NOTE (mristin):
{I}// We are aware that exceptions in constructors should be avoided to prevent
{I}// bug related to uninitialized object state. However, in this case, we do not
{I}// see any risk for such a mistake.
{I}if (ranges.empty()) {{
{II}throw std::invalid_argument(
{III}"Unexpected NotSet instruction with empty ranges"
{II});
{I}}}

{I}for (size_t i = 1; i < ranges.size(); ++i) {{
{II}if (ranges[i - 1].last >= ranges[i].first) {{
{III}throw std::invalid_argument(
{IIII}common::Concat(
{IIIII}"The ranges for an InstructionSet are unexpectedly either "
{IIIII}"not sorted or overlapping. The range at index ",
{IIIII}std::to_string(i - 1),
{IIIII}" is ",
{IIIII}to_string(ranges[i - 1]),
{IIIII}" and the range at index ",
{IIIII}std::to_string(i),
{IIIII}" is ",
{IIIII}to_string(ranges[i])
{IIII})
{III});
{II}}}
{I}}}
}}

InstructionKind InstructionSet::kind() const {{
{I}return InstructionKind::Set;
}}

std::string to_string(const InstructionSet& instruction) {{
{I}std::stringstream ss;
{I}ss << "set '";

{I}for (const auto& range : instruction.ranges) {{
{II}ss << to_string(range);
{I}}}

{I}ss << "'";
{I}return ss.str();
}}

InstructionNotSet::InstructionNotSet(
{I}std::vector<Range> a_ranges
) :
{I}ranges(std::move(a_ranges)) {{
{I}// NOTE (mristin):
{I}// We are aware that exceptions in constructors should be avoided to prevent
{I}// bug related to uninitialized object state. However, in this case, we do not
{I}// see any risk for such a mistake.
{I}if (ranges.empty()) {{
{II}throw std::invalid_argument(
{III}"Unexpected NotSet instruction with empty ranges"
{II});
{I}}}

{I}for (size_t i = 1; i < ranges.size(); ++i) {{
{II}if (ranges[i - 1].last >= ranges[i].first) {{
{III}throw std::invalid_argument(
{IIII}common::Concat(
{IIIII}"The ranges for an InstructionNotSet are unexpectedly either "
{IIIII}"not sorted or overlapping. The range at index ",
{IIIII}std::to_string(i - 1),
{IIIII}" is ",
{IIIII}to_string(ranges[i - 1]),
{IIIII}" and the range at index ",
{IIIII}std::to_string(i),
{IIIII}" is ",
{IIIII}to_string(ranges[i])
{IIII})
{III});
{II}}}
{I}}}
}}

InstructionKind InstructionNotSet::kind() const {{
{I}return InstructionKind::NotSet;
}}

std::string to_string(const InstructionNotSet& instruction) {{
{I}std::stringstream ss;
{I}ss << "not-set '";

{I}for (const auto& range : instruction.ranges) {{
{II}ss << to_string(range);
{I}}}

{I}ss << "'";
{I}return ss.str();
}}

InstructionKind InstructionAny::kind() const {{
{I}return InstructionKind::Any;
}}

std::string to_string(const InstructionAny&) {{
{I}return "any";
}}

InstructionKind InstructionMatch::kind() const {{
{I}return InstructionKind::Match;
}}

std::string to_string(const InstructionMatch&) {{
{I}return "match";
}}

InstructionJump::InstructionJump(
{I}size_t a_target
) :
{I}target(a_target) {{
{I}// Intentionally empty.
}}

InstructionKind InstructionJump::kind() const {{
{I}return InstructionKind::Jump;
}}

std::string to_string(const InstructionJump& instruction) {{
{I}return common::Concat(
{II}"jump ",
{II}std::to_string(instruction.target)
{I});
}}

InstructionSplit::InstructionSplit(
{I}size_t a_first_target,
{I}size_t a_second_target
) :
{I}first_target(a_first_target),
{I}second_target(a_second_target) {{
{I}// Intentionally empty.
}}

InstructionKind InstructionSplit::kind() const {{
{I}return InstructionKind::Split;
}}

std::string to_string(const InstructionSplit& instruction) {{
{I}return common::Concat(
{II}"split ",
{III}std::to_string(instruction.first_target),
{III}", ",
{III}std::to_string(instruction.second_target)
{I});
}}

InstructionKind InstructionEnd::kind() const {{
{I}return InstructionKind::End;
}}

std::string to_string(const InstructionEnd&) {{
{I}return "end";
}}

std::string to_string(const Instruction& instruction) {{
{I}switch (instruction.kind()) {{
{II}case InstructionKind::Char:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionChar&
{IIII}>(instruction)
{III});

{II}case InstructionKind::Set:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionSet&
{IIII}>(instruction)
{III});

{II}case InstructionKind::NotSet:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionNotSet&
{IIII}>(instruction)
{III});

{II}case InstructionKind::Any:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionAny&
{IIII}>(instruction)
{III});

{II}case InstructionKind::Match:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionMatch&
{IIII}>(instruction)
{III});

{II}case InstructionKind::Jump:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionJump&
{IIII}>(instruction)
{III});

{II}case InstructionKind::Split:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionSplit&
{IIII}>(instruction)
{III});

{II}case InstructionKind::End:
{III}return to_string(
{IIII}static_cast<   // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIII}const InstructionEnd&
{IIII}>(instruction)
{III});

{II}default:
{III}throw std::logic_error(
{IIII}common::Concat(
{IIIII}"Unhandled instruction kind: ",
{IIIII}std::to_string(
{IIIIII}static_cast<std::uint8_t>(instruction.kind())
{IIIII})
{IIII})
{III});
{I}}}
}}

std::string to_string(
{I}const std::vector<std::unique_ptr<Instruction> >& instructions
) {{
{I}std::stringstream ss;
{I}for (size_t i = 0; i < instructions.size(); ++i) {{
{II}ss
{III}<< std::setw(4) << i << ": "
{III}<< to_string(*instructions[i]) << std::endl;
{I}}}

{I}return ss.str();
}}

bool CharacterInRanges(
{I}const std::vector<Range>& ranges,
{I}wchar_t character
) {{
{I}if (ranges.empty()) {{
{II}return false;
{I}}}

{I}if (ranges.size() == 1) {{
{II}return (ranges[0].first <= character && character <= ranges[0].last);
{I}}}

{I}// Binary search
{I}size_t begin = 0;
{I}size_t end = ranges.size();

{I}while (true) {{
{II}if (begin == end) {{
{III}return false;
{II}}}

{II}// NOTE (mristin):
{II}// Most implementations of the binary search are buggy, see:
{II}// https://en.wikipedia.org/wiki/Binary_search_algorithm#Implementation_issues.
{II}//
{II}// We try to avert some of the bugs by explicitly handling the case where there
{II}// are at most 3 elements in the segment, so we stop here instead of proceeding
{II}// recursively.
{II}if (end - begin <= 3) {{
{III}for (size_t i = begin; i < end; ++i) {{
{IIII}const Range& range = ranges[i];
{IIII}if (range.first <= character && character <= range.last) {{
{IIIII}return true;
{IIII}}}
{III}}}
{III}return false;
{II}}}

{II}const size_t middle = (begin + end) / 2;
{II}const Range& range = ranges[middle];
{II}if (character < range.first) {{
{III}end = middle;
{II}}} else if (character > range.last) {{
{III}begin = middle;
{II}}} else {{
{III}return true;
{II}}}
{I}}}
}}

/**
 * Keep track of the threads currently being executed.
 */
class ThreadList {{
 public:
{I}explicit ThreadList(size_t program_size) {{
{II}has_.resize(program_size, false);
{II}items_.reserve(program_size);
{I}}}

{I}/**
{I} * Add a new thread for the given program counter if it is not already in the list.
{I} */
{I}void Spawn(size_t program_counter) {{
{II}#ifdef DEBUG
{II}if (program_counter >= program_size_) {{
{III}throw std::invalid_argument(
{IIII}common::Concat(
{IIIII}"Unexpected spawning of a thread at the program counter ",
{IIIII}std::to_string(program_counter),
{IIIII}" since the program size was indicated to be ",
{IIIII}std::to_string(program_size_)
{IIII})
{III});
{II}}}
{II}#endif

{II}if (has_[program_counter]) {{
{III}return;
{II}}}

{II}has_[program_counter] = true;
{II}items_.push_back(program_counter);
{I}}}

{I}bool Empty() const {{
{II}return items_.empty();
{I}}}

{I}/**
{I} * Pop the thread from the back, returning its program counter.
{I} *
{I} * The order of the threads is not guaranteed.
{I} */
{I}size_t Pop() {{
{II}#ifdef DEBUG
{II}if (items_.empty()) {{
{III}throw std::logic_error(
{IIII}"You tried to pop from an empty thread list."
{IIII});
{II}}}
{II}#endif

{II}const size_t program_counter = items_.back();
{II}items_.pop_back();
{II}has_[program_counter] = false;
{II}return program_counter;
{I}}}

{I}/**
{I} * Clear the thread list, keeping its memory capacity.
{I} */
{I}void Clear() {{
{II}std::fill(has_.begin(), has_.end(), false);
{II}items_.clear();
{I}}}

{I}/**
{I} * Return the program counters corresponding to the spawned threads.
{I} */
{I}const std::vector<size_t>& Items() const {{
{II}return items_;
{I}}};

 private:
{I}/**
{I} * Keep track of the program counters corresponding to the threads so that we can
{I} * avoid the duplicate threads.
{I} */
{I}std::vector<bool> has_;

{I}/**
{I} * Keep track of the active threads.
{I} */
{I}std::vector<size_t> items_;
}};

std::string to_string(const ThreadList& thread_list) {{
{I}if (thread_list.Empty()) {{
{II}return "[]";
{I}}}

{I}std::vector<size_t> items(thread_list.Items());
{I}std::sort(items.begin(), items.end());

{I}std::stringstream ss;
{I}ss << "[";
{I}ss << std::to_string(items[0]);

{I}for (size_t i = 1; i < items.size(); ++i) {{
{II}ss << ", " << items[i];
{I}}}
{I}ss << "]";

{I}return ss.str();
}}

/**
 * Try to match the program against the text.
 *
 * @return true if the text matches
 */
bool Match(
{I}const std::vector<std::unique_ptr<Instruction> >& program,
{I}const std::wstring& text
) {{
{I}if (program.empty()) {{
{II}return false;
{I}}}

{I}// NOTE (mristin):
{I}// We validate at the beginning so that we can avoid checks in
{I}// the instruction loops.
{I}for (size_t i = 0; i < program.size(); ++i) {{
{II}const Instruction& instruction = *program[i];

{II}switch (instruction.kind()) {{
{III}case InstructionKind::Jump: {{
{IIII}const auto& instruction_jump(
{IIIII}static_cast<  // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIII}const InstructionJump&
{IIIII}>(
{IIIIII}instruction
{IIIII})
{IIII});

{IIII}if (instruction_jump.target >= program.size()) {{
{IIIII}throw std::invalid_argument(
{IIIIII}common::Concat(
{IIIIIII}"Unexpected jump beyond the end of the program. Program has ",
{IIIIIII}std::to_string(program.size()),
{IIIIIII}" instruction(s), but the instruction ",
{IIIIIII}std::to_string(i),
{IIIIIII}" wants to jump to ",
{IIIIIII}std::to_string(instruction_jump.target)
{IIIIII})
{IIIII});
{IIII}}}
{IIII}break;
{III}}}

{III}case InstructionKind::Split: {{
{IIII}const auto& instruction_split(
{IIIII}static_cast<  // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIII}const InstructionSplit&
{IIIII}>(
{IIIIII}instruction
{IIIII})
{IIII});

{IIII}if (instruction_split.first_target >= program.size()) {{
{IIIII}throw std::invalid_argument(
{IIIIII}common::Concat(
{IIIIIII}"Unexpected split & jump beyond the end of the program. Program has ",
{IIIIIII}std::to_string(program.size()),
{IIIIIII}" instruction(s), but the instruction ",
{IIIIIII}std::to_string(i),
{IIIIIII}" wants to split and make the first jump to ",
{IIIIIII}std::to_string(instruction_split.first_target)
{IIIIII})
{IIIII});
{IIII}}}

{IIII}if (instruction_split.second_target >= program.size()) {{
{IIIII}throw std::invalid_argument(
{IIIIII}common::Concat(
{IIIIIII}"Unexpected split & jump beyond the end of the program. Program has ",
{IIIIIII}std::to_string(program.size()),
{IIIIIII}" instruction(s), but the instruction ",
{IIIIIII}std::to_string(i),
{IIIIIII}" wants to split and make the second jump to ",
{IIIIIII}std::to_string(instruction_split.second_target)
{IIIIII})
{IIIII});
{IIII}}}

{IIII}break;
{III}}}

{III}default:
{IIII}continue;
{II}}}
{I}}}

{I}// NOTE (mristin):
{I}// See: https://swtch.com/~rsc/regexp/regexp2.html,
{I}// Section "Thompson's Implementation".

{I}std::unique_ptr<ThreadList> clist(common::make_unique<ThreadList>(program.size()));
{I}std::unique_ptr<ThreadList> nlist(common::make_unique<ThreadList>(program.size()));

{I}clist->Spawn(0);

{I}for (const wchar_t character : text) {{
{II}#ifdef DEBUG
{II}if (!nlist->Empty()) {{
{III}throw std::logic_error(
{IIII}"Expected the list of next-to-be-executed threads to be empty, "
{IIII}"but it was not."
{III});
{II}}}
{II}#endif

{II}while (!clist->Empty()) {{
{III}const size_t program_counter = clist->Pop();

{III}#ifdef DEBUG
{III}if (program_counter >= program.size()) {{
{IIII}throw std::logic_error(
{IIIII}common::Concat(
{IIIIII}"Unexpected program counter beyond the program. The program size was ",
{IIIIII}std::to_string(program.size()),
{IIIIII}", while the program counter of a thread was ",
{IIIIII}std::to_string(program_counter)
{IIIII})
{IIII});
{III}}}
{III}#endif

{III}const Instruction& instruction = *program[program_counter];

{III}switch (instruction.kind()) {{
{IIII}case InstructionKind::Char: {{
{IIIII}const auto& instruction_char(
{IIIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIIII}const InstructionChar&
{IIIIII}>(instruction)
{IIIII});

{IIIII}if (character != instruction_char.character) {{
{IIIIII}// The matching failed for this thread.
{IIIIII}break;
{IIIII}}}

{IIIII}nlist->Spawn(program_counter + 1);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::Set: {{
{IIIII}const auto& instruction_set(
{IIIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIIII}const InstructionSet&
{IIIIII}>(instruction)
{IIIII});

{IIIII}if (!CharacterInRanges(instruction_set.ranges, character)) {{
{IIIIII}// The matching failed for this thread.
{IIIIII}break;
{IIIII}}}

{IIIII}nlist->Spawn(program_counter + 1);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::NotSet: {{
{IIIII}const auto& instruction_not_set(
{IIIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIIII}const InstructionNotSet&
{IIIIII}>(instruction)
{IIIII});

{IIIII}if (CharacterInRanges(instruction_not_set.ranges, character)) {{
{IIIIII}// The matching failed for this thread.
{IIIIII}break;
{IIIII}}}

{IIIII}nlist->Spawn(program_counter + 1);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::Any: {{
{IIIII}// NOTE (mristin):
{IIIII}// We simply proceed to the next instruction at the next character without
{IIIII}// any checks.
{IIIII}nlist->Spawn(program_counter + 1);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::Match:
{IIIII}return true;

{IIII}case InstructionKind::Jump: {{
{IIIII}const auto& instruction_jump(
{IIIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIIII}const InstructionJump&
{IIIIII}>(instruction)
{IIIII});

{IIIII}clist->Spawn(instruction_jump.target);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::Split: {{
{IIIII}const auto& instruction_split(
{IIIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIIII}const InstructionSplit&
{IIIIII}>(instruction)
{IIIII});

{IIIII}clist->Spawn(instruction_split.first_target);
{IIIII}clist->Spawn(instruction_split.second_target);
{IIIII}break;
{IIII}}}

{IIII}case InstructionKind::End: {{
{IIIII}// The matching failed for this thread as we have just consumed
{IIIII}// a character.
{IIIII}break;
{IIII}}}

{IIII}default:
{IIIII}throw std::logic_error(
{IIIIII}common::Concat(
{IIIIIII}"Unhandled instruction kind: ",
{IIIIIII}std::to_string(
{IIIIIIII}static_cast<std::uint8_t>(instruction.kind())
{IIIIIII})
{IIIIII})
{IIIII});
{III}}}
{II}}}

{II}std::swap(clist, nlist);
{II}nlist->Clear();
{I}}}

{I}// NOTE (mristin):
{I}// We need to process any pending jumps, splits and matches even tough there are
{I}// no more characters to consume.
{I}while (!clist->Empty()) {{
{II}const size_t program_counter = clist->Pop();

{II}#ifdef DEBUG
{II}if (program_counter >= program.size()) {{
{III}throw std::logic_error(
{IIII}common::Concat(
{IIIII}"Unexpected program counter beyond the program. The program size was ",
{IIIII}std::to_string(program.size()),
{IIIII}", while the program counter of a thread was ",
{IIIII}std::to_string(program_counter)
{IIII})
{III});
{II}}}
{II}#endif

{II}const Instruction& instruction = *program[program_counter];

{II}switch (instruction.kind()) {{
{III}case InstructionKind::Char: {{ // NOLINT(bugprone-branch-clone)
{IIII}// We reached the end-of-input so there are no characters to be matched.
{IIII}// This thread needs therefore to die.
{IIII}break;
{III}}}

{III}case InstructionKind::Set: {{
{IIII}// We reached the end-of-input so there are no character sets to be matched.
{IIII}// This thread needs therefore to die.
{IIII}break;
{III}}}

{III}case InstructionKind::NotSet: {{
{IIII}// We reached the end-of-input so there are no character sets to be matched.
{IIII}// This thread needs therefore to die.
{IIII}break;
{III}}}

{III}case InstructionKind::Any: {{
{IIII}// We reached the end-of-input so there are no characters to be matched.
{IIII}// This thread needs therefore to die.
{IIII}break;
{III}}}

{III}case InstructionKind::Match:
{IIII}return true;

{III}case InstructionKind::Jump: {{
{IIII}const auto& instruction_jump(
{IIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIII}const InstructionJump&
{IIIII}>(instruction)
{IIII});

{IIII}clist->Spawn(instruction_jump.target);
{IIII}break;
{III}}}

{III}case InstructionKind::Split: {{
{IIII}const auto& instruction_split(
{IIIII}static_cast< // NOLINT(cppcoreguidelines-pro-type-static-cast-downcast)
{IIIIII}const InstructionSplit&
{IIIII}>(instruction)
{IIII});

{IIII}clist->Spawn(instruction_split.first_target);
{IIII}clist->Spawn(instruction_split.second_target);
{IIII}break;
{III}}}

{III}case InstructionKind::End: {{
{IIII}// We reached the end-of-input so we match and move to the next instruction.
{IIII}clist->Spawn(program_counter + 1);
{IIII}break;
{III}}}

{III}default:
{IIII}throw std::logic_error(
{IIIII}common::Concat(
{IIIIII}"Unhandled instruction kind: ",
{IIIIII}std::to_string(
{IIIIIII}static_cast<std::uint8_t>(instruction.kind())
{IIIIII})
{IIIII})
{IIII});
{II}}}
{I}}}

{I}return false;
}}"""
        ),
        cpp_common.generate_namespace_closing(namespace),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Run a smoke-test as a preliminary test that a meta-model is ready for generation."""

import argparse
import pathlib
import sys
from typing import TextIO, List, MutableMapping

import aas_core_codegen
import aas_core_codegen.smoke
from aas_core_codegen import (
    parse,
    run,
    intermediate,
    infer_for_schema,
    specific_implementations,
)
from aas_core_codegen.common import LinenoColumner, Error, Stripped
from aas_core_codegen.csharp import common as csharp_common, lib as csharp_lib

assert __doc__ == aas_core_codegen.smoke.__doc__


def _smoke_transpile_to_csharp(symbol_table: intermediate.SymbolTable) -> List[Error]:
    """Try to transpile to C# what you can without snippets."""
    # fmt: off
    verified_symbol_table, errors_in_types = (
        csharp_lib.verify_for_types(
            symbol_table
        )
    )
    # fmt: on
    if errors_in_types is not None:
        return errors_in_types

    assert verified_symbol_table is not None

    errors = []

    spec_impls = (
        dict()
    )  # type: MutableMapping[specific_implementations.ImplementationKey, Stripped]

    dummy_implementation = Stripped("DUMMY IMPLEMENTATION")

    for cls in symbol_table.classes:
        if cls.is_implementation_specific:
            key = specific_implementations.ImplementationKey(
                f"Types/{cls.name}/{cls.name}.cs"
            )
            spec_impls[key] = dummy_implementation
            continue

        for method in cls.methods:
            if isinstance(method, intermediate.ImplementationSpecificMethod):
                key = specific_implementations.ImplementationKey(
                    f"Types/{cls.name}/{method.name}.cs"
                )
                spec_impls[key] = dummy_implementation

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.cs"
            )
            spec_impls[key] = dummy_implementation

    namespace = csharp_common.NamespaceIdentifier("DummyNamespace")

    _, types_errors = csharp_lib.generate_types(
        symbol_table=verified_symbol_table, namespace=namespace, spec_impls=spec_impls
    )
    if types_errors is not None:
        errors.extend(types_errors)

    _, verification_errors = csharp_lib.generate_verification(
        symbol_table=symbol_table, namespace=namespace, spec_impls=spec_impls
    )
    if verification_errors is not None:
        errors.extend(verification_errors)

    return errors


def execute(model_path: pathlib.Path, stderr: TextIO) -> int:
    """Run the smoke test."""
    text = model_path.read_text(encoding="utf-8")

    atok, parse_exception = parse.source_to_atok(source=text)
    if parse_exception:
        if isinstance(parse_exception, SyntaxError):
            stderr.write(
                f"Failed to parse the meta-model {model_path}: "
                f"invalid syntax at line {parse_exception.lineno}\n"
            )
        else:
            stderr.write(
                f"Failed to parse the meta-model {model_path}: {parse_exception}\n"
            )

        return 1

    assert atok is not None

    import_errors = parse.check_expected_imports(atok=atok)
    if import_errors:
        run.write_error_report(
            message="One or more unexpected imports in the meta-model",
            errors=import_errors,
            stderr=stderr,
        )

        return 1

    lineno_columner = LinenoColumner(atok=atok)

    parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    if error is not None:
        run.write_error_report(
            message=f"Failed to construct the symbol table from {model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    assert parsed_symbol_table is not None

    ir_symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table,
        atok=atok,
    )
    if error is not None:
        run.write_error_report(
            message=f"Failed to translate the parsed symbol table "
            f"to intermediate symbol table "
            f"based on {model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    assert ir_symbol_table is not None

    constraints_by_class, errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=ir_symbol_table
    )
    if errors is not None:
        run.write_error_report(
            message=f"Failed to infer the constraints by class for the schemas "
            f"based on {model_path}",
            errors=[lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )

        return 1

    assert constraints_by_class is not None

    _, error = infer_for_schema.merge_constraints_with_ancestors(
        symbol_table=ir_symbol_table, constraints_by_class=constraints_by_class
    )
    if errors is not None:
        run.write_error_report(
            message=f"Failed to merge the constraints with ancestors for the schemas "
            f"based on {model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    errors = _smoke_transpile_to_csharp(symbol_table=ir_symbol_table)

    if len(errors) > 0:
        run.write_error_report(
            message=f"Failed to smoke-transpile what we could without snippets to C# "
            f"based on {model_path}",
            errors=[lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )

        return 1

    return 0


def main(prog: str) -> int:
    """Execute the main routine."""
    # NOTE (mristin, 2022-03-28):
    # The module ``argparse`` is not flexible enough to understand special options such
    # as ``--version`` so we manually hard-wire.
    if "--version" in sys.argv and "--help" not in sys.argv:
        print(aas_core_codegen.__version__)
        return 0

    parser = argparse.ArgumentParser(prog=prog, description=__doc__)
    parser.add_argument("--model_path", help="path to the meta-model", required=True)
    parser.add_argument(
        "--version", help="show the current version and exit", action="store_true"
    )
    args = parser.parse_args()

    return execute(model_path=pathlib.Path(args.model_path), stderr=sys.stderr)


def entry_point() -> int:
    """Provide an entry point for a console script."""
    return main(prog="aas-core-codegen-smoke")


if __name__ == "__main__":
    sys.exit(main(prog="aas-core-codegen-smoke"))



"""Generate the TypeScript data structures from the intermediate representation."""
import io
import textwrap
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    cast,
    Union,
    Mapping,
    Final,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen import specific_implementations
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.intermediate import (
    construction as intermediate_construction,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
    description as typescript_description,
    unrolling as typescript_unrolling,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# region Checks


def _human_readable_identifier(
    something: Union[
        intermediate.Enumeration, intermediate.AbstractClass, intermediate.ConcreteClass
    ]
) -> str:
    """
    Represent ``something`` in a human-readable text.

    The reader should be able to trace ``something`` back to the meta-model.
    """
    # NOTE (mristin, 2022-11-10):
    # This function has been copy-pasted from
    # :py:mod:`aas_core_codegen.python.structure._generate`. We tried to refactor it to
    # :py:mod:`aas_core_codegen.intermediate`, but it turned out that the refactored
    # code was nigh unreadable. So we preferred a little bit of copying to a little
    # bit of complexity.

    result: str

    if isinstance(something, intermediate.Enumeration):
        result = f"meta-model enumeration {something.name!r}"
    elif isinstance(something, intermediate.AbstractClass):
        result = f"meta-model abstract class {something.name!r}"
    elif isinstance(something, intermediate.ConcreteClass):
        result = f"meta-model concrete class {something.name!r}"
    else:
        assert_never(something)

    return result


def _verify_intra_structure_collisions(
    our_type: intermediate.OurType,
) -> Optional[Error]:
    """Verify that no member names collide in the TypeScript structure of our type."""
    errors = []  # type: List[Error]

    if isinstance(our_type, intermediate.Enumeration):
        enum_literal_map = (
            dict()
        )  # type: Dict[Identifier, intermediate.EnumerationLiteral]

        for literal in our_type.literals:
            literal_name = typescript_naming.enum_literal_name(literal.name)
            colliding_literal = enum_literal_map.get(literal_name, None)
            if colliding_literal is not None:
                errors.append(
                    Error(
                        literal.parsed.node,
                        f"The TypeScript name, {literal_name!r}, "
                        f"for the literal {literal.name!r} collides with "
                        f"the TypeScript name of another "
                        f"literal {colliding_literal.name!r}",
                    )
                )
            else:
                enum_literal_map[literal_name] = literal

    elif isinstance(our_type, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(our_type, intermediate.Class):
        observed_member_names = {}  # type: Dict[Identifier, str]

        for prop in our_type.properties:
            prop_name = typescript_naming.property_name(prop.name)
            if prop_name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"TypeScript property {prop_name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[prop_name]}",
                    )
                )
            else:
                observed_member_names[prop_name] = (
                    f"TypeScript property {prop_name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

        for method in our_type.methods:
            method_name = typescript_naming.method_name(method.name)

            if method_name in observed_member_names:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"TypeScript method {method_name!r} corresponding "
                        f"to the meta-model method {method.name!r} collides with "
                        f"the {observed_member_names[method_name]}",
                    )
                )
            else:
                observed_member_names[method_name] = (
                    f"TypeScript method {method_name!r} corresponding to "
                    f"the meta-model method {method.name!r}"
                )

    else:
        assert_never(our_type)

    if len(errors) > 0:
        errors.append(
            Error(
                our_type.parsed.node,
                f"Naming collision(s) in TypeScript code for our type {our_type.name!r}",
                underlying=errors,
            )
        )

    return None


def _verify_structure_name_collisions(
    symbol_table: intermediate.SymbolTable,
) -> List[Error]:
    """Verify that the TypeScript names of the structures do not collide."""
    observed_structure_names: Dict[
        Identifier,
        Union[
            intermediate.Enumeration,
            intermediate.AbstractClass,
            intermediate.ConcreteClass,
        ],
    ] = dict()

    errors = []  # type: List[Error]

    # region Inter-structure collisions

    for our_type in symbol_table.our_types:
        if not isinstance(
            our_type,
            (
                intermediate.Enumeration,
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            continue

        name = typescript_naming.name_of(our_type)

        other = observed_structure_names.get(name, None)

        if other is not None:
            errors.append(
                Error(
                    our_type.parsed.node,
                    f"The TypeScript name {name!r} "
                    f"of the "
                    f"{_human_readable_identifier(our_type)} "
                    f"collides with the TypeScript name "
                    f"of the "
                    f"{_human_readable_identifier(other)}",
                )
            )
        else:
            observed_structure_names[name] = our_type

    # endregion

    # region Intra-structure collisions

    for our_type in symbol_table.our_types:
        collision_error = _verify_intra_structure_collisions(our_type=our_type)

        if collision_error is not None:
            errors.append(collision_error)

    # endregion

    return errors


class VerifiedIntermediateSymbolTable(intermediate.SymbolTable):
    """Represent a verified symbol table which can be used for code generation."""

    # noinspection PyInitNewSignature
    def __new__(
        cls, symbol_table: intermediate.SymbolTable
    ) -> "VerifiedIntermediateSymbolTable":
        raise AssertionError("Only for type annotation")


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def verify(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[VerifiedIntermediateSymbolTable], Optional[List[Error]]]:
    """Verify that TypeScript code can be generated from the ``symbol_table``."""
    errors = []  # type: List[Error]

    structure_name_collisions = _verify_structure_name_collisions(
        symbol_table=symbol_table
    )

    errors.extend(structure_name_collisions)

    if len(errors) > 0:
        return None, errors

    return cast(VerifiedIntermediateSymbolTable, symbol_table), None


# endregion

# region Generation


def _generate_model_type_enum(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the enumerator to represent runtime model type of an instance."""
    blocks = []

    for i, cls in enumerate(symbol_table.concrete_classes):
        literal = typescript_naming.enum_literal_name(cls.name)

        blocks.append(Stripped(f"{literal} = {i}"))

    enum_name = typescript_naming.enum_name(Identifier("Model_type"))

    body = ",\n".join(blocks)

    return Stripped(
        f"""\
/**
 * Represent runtime model type of an instance.
 */
export enum {enum_name} {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_over_model_type_enum(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the iterator over the model type enumerator."""
    blocks = []

    model_type_enum = typescript_naming.enum_name(Identifier("Model_type"))

    for i, cls in enumerate(symbol_table.concrete_classes):
        literal = typescript_naming.enum_literal_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
yield <{model_type_enum}>{i};  // {literal}"""
            )
        )

    over_model_type = typescript_naming.function_name(Identifier("over_model_type"))

    body = "\n".join(blocks)

    return Stripped(
        f"""\
/**
 * Iterate over the literals of {{@link {model_type_enum}}}.
 *
 * @remark
 * TypeScript does not provide an elegant way to iterate over the literals, so
 * this function helps you avoid common errors and pitfalls.
 *
 * @return iterator over the literals
 */
export function *{over_model_type} (
): Iterable<{model_type_enum}> {{
{I}// NOTE (mristin, 2022-12-03):
{I}// We yield numbers instead of literals to avoid name lookups on platforms
{I}// which do not provide JIT compilation of hot paths.
{I}{indent_but_first_line(body, I)}
}}"""
    )


@require(lambda enumeration, literal: id(literal) in enumeration.literal_id_set)
@require(lambda literal: literal.description is not None)
def _generate_comment_for_enumeration_literal(
    enumeration: intermediate.Enumeration,
    literal: intermediate.EnumerationLiteral,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    # NOTE (mristin, 2022-10-29):
    # We need to state the pre-condition for the second time for mypy.
    assert literal.description is not None

    # fmt: off
    comment, errors = (
        typescript_description.generate_documentation_comment_for_summary_remarks(
            description=literal.description,
            context=typescript_description.Context(
                module=typescript_common.TYPES_MODULE,
                cls_or_enum=enumeration
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@require(lambda cls_or_enum: cls_or_enum.description is not None)
def _generate_comment_for_cls_or_enum(
    cls_or_enum: Union[intermediate.Enumeration, intermediate.ClassUnion],
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for our type."""
    # NOTE (mristin, 2022-10-29):
    # We need to state the pre-condition for the second time for mypy.
    assert cls_or_enum.description is not None

    # fmt: off
    comment, errors = (
        typescript_description
        .generate_documentation_comment_for_summary_remarks_constraints(
            description=cls_or_enum.description,
            context=typescript_description.Context(
                module=typescript_common.TYPES_MODULE, cls_or_enum=None
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_enum(
    enum: intermediate.Enumeration,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the TypeScript code for the enum."""
    writer = io.StringIO()

    errors = []  # type: List[Error]

    comment = None  # type: Optional[Stripped]
    if enum.description is not None:
        comment, comment_errors = _generate_comment_for_cls_or_enum(cls_or_enum=enum)

        if comment_errors:
            errors.append(
                Error(
                    enum.description.parsed.node,
                    f"Failed to generate the comment "
                    f"for the enumeration {enum.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None

    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    name = typescript_naming.enum_name(enum.name)

    if len(enum.literals) == 0:
        writer.write(f"export enum {name} {{}}")
    else:
        writer.write(f"export enum {name} {{\n")

        for i, literal in enumerate(enum.literals):
            if literal.description is not None:
                comment, comment_errors = _generate_comment_for_enumeration_literal(
                    enumeration=enum, literal=literal
                )
                if comment_errors is not None:
                    errors.append(
                        Error(
                            literal.description.parsed.node,
                            f"Failed to generate the documentation comment "
                            f"for enumeration literal {literal.name!r}",
                            comment_errors,
                        )
                    )
                else:
                    assert comment is not None
                    writer.write(textwrap.indent(comment, I))
                    writer.write("\n")

            literal_name = typescript_naming.enum_literal_name(literal.name)

            # NOTE (mristin, 2022-11-10):
            # We optimize for comparisons instead of stringification.
            # The stringification is delegated to a separate module, with efficient
            # array look-ups to get the actual string value of a literal.

            if i == 0:
                writer.write(textwrap.indent(f"{literal_name} = 0", I))
            else:
                writer.write(textwrap.indent(f"{literal_name}", I))

            if i < len(enum.literals) - 1:
                writer.write(",")

            writer.write("\n")

        writer.write("}")

    if len(errors) > 0:
        return None, Error(
            enum.parsed.node,
            f"Failed to generate the TypeScript code for the enumeration {enum.name!r}",
            errors,
        )

    return Stripped(writer.getvalue()), None


def _generate_over_enum(enum: intermediate.Enumeration) -> Stripped:
    """Generate the TypeScript code for the function to iterate over the literals."""
    name = typescript_naming.enum_name(enum.name)
    function_name = typescript_naming.function_name(Identifier(f"over_{enum.name}"))

    statements = [
        f"yield <{name}>{i}; // {typescript_naming.enum_literal_name(literal.name)}"
        for i, literal in enumerate(enum.literals)
    ]
    body = "\n".join(statements)

    return Stripped(
        f"""\
/**
 * Iterate over the literals of {{@link {name}}}.
 *
 * @remark
 * TypeScript does not provide an elegant way to iterate over the literals, so
 * this function helps you avoid common errors and pitfalls.
 *
 * @return iterator over the literals
 */
export function *{function_name}(
): IterableIterator<{name}> {{
{I}// NOTE (mristin, 2022-12-03):
{I}// We yield numbers instead of literals to avoid name lookups on platforms
{I}// which do not provide JIT compilation of hot paths.
{I}{indent_but_first_line(body, I)}
}}"""
    )


class _DescendBodyUnroller(typescript_unrolling.AbstractUnroller):
    """Generate the code that unrolls descent into an element."""

    #: If set, generates the code with unrolled yields.
    #: Otherwise, we do not unroll recursively.
    _recurse: Final[bool]

    #: Pre-computed descendability map. A type is descendable if we should unroll it
    #: further.
    _descendability: Final[Mapping[intermediate.TypeAnnotationUnion, bool]]

    #: Generator of loop variable names.
    #:
    #: We generate for each list iteration a new variable to avoid shadowing
    #: of the variables. Even though TypeScript tracks variables in the block scope,
    #: we want to avoid confusion for the reader.
    _generator_for_loop_variables: Final[typescript_common.GeneratorForLoopVariables]

    def __init__(
        self,
        recurse: bool,
        descendability: Mapping[intermediate.TypeAnnotationUnion, bool],
        generator_for_loop_variables: typescript_common.GeneratorForLoopVariables,
    ) -> None:
        """Initialize with the given values."""
        self._recurse = recurse
        self._descendability = descendability
        self._generator_for_loop_variables = generator_for_loop_variables

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[typescript_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        # We can not descend into a primitive type.
        return []

    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[typescript_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            # We can not descend into an enumeration.
            return []

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # We can not descend into a primitive type.
            return []

        assert isinstance(our_type, intermediate.Class)  # Exhaustively match

        result = [typescript_unrolling.Node(f"yield {unrollee_expr};", children=[])]

        if self._recurse:
            if self._descendability[type_annotation]:
                result.append(
                    typescript_unrolling.Node(
                        text=f"yield * {unrollee_expr}.descend();",
                        children=[],
                    )
                )

        return result

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[typescript_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        if (
            not self._recurse
            and isinstance(type_annotation.items, intermediate.OurTypeAnnotation)
            and isinstance(
                type_annotation.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
        ):
            return [typescript_unrolling.Node(f"yield * {unrollee_expr};", children=[])]

        loop_var = next(self._generator_for_loop_variables)
        children = self.unroll(
            unrollee_expr=loop_var,
            type_annotation=type_annotation.items,
            path=[],  # Path is unused in this context
            list_loop_level=list_loop_level + 1,
        )

        if len(children) == 0:
            return []

        node = typescript_unrolling.Node(
            text=f"for (const {loop_var} of {unrollee_expr})",
            children=children,
        )

        return [node]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[typescript_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            list_loop_level=list_loop_level,
        )

        if len(children) == 0:
            return []

        return [
            typescript_unrolling.Node(
                text=f"if ({unrollee_expr} !== null)", children=children
            )
        ]


def _generate_descend_body(cls: intermediate.ConcreteClass, recurse: bool) -> Stripped:
    """
    Generate the body of the ``descend`` and ``descend_once`` methods.

    With this function, we unroll the recursion as a simple optimization
    in the recursive case.
    """
    blocks = []  # type: List[Stripped]

    generator_for_loop_variables = typescript_common.GeneratorForLoopVariables()

    for prop in cls.properties:
        descendability = intermediate.map_descendability(
            type_annotation=prop.type_annotation
        )

        if not descendability[prop.type_annotation]:
            continue

        # region Unroll

        unroller = _DescendBodyUnroller(
            recurse=recurse,
            descendability=descendability,
            generator_for_loop_variables=generator_for_loop_variables,
        )

        roots = unroller.unroll(
            unrollee_expr=f"this.{typescript_naming.property_name(prop.name)}",
            type_annotation=prop.type_annotation,
            path=[],  # We do not use path in this context
            list_loop_level=0,
        )

        assert len(roots) > 0, (
            "Since the type annotation was descendable, we must have obtained "
            "at least one unrolling node"
        )

        blocks.extend(Stripped(typescript_unrolling.render(root)) for root in roots)

        # endregion

    if len(blocks) == 0:
        blocks.append(Stripped("// No descendable properties"))

    return Stripped("\n\n".join(blocks))


def _generate_descend_once_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the ``descend_once`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=False)

    return Stripped(
        f"""\
/**
 * Iterate over the instances referenced from this instance.
 *
 * We do not recurse into the referenced instances.
 *
 * @returns Iterator over the referenced instances
 */
*descendOnce(): IterableIterator<Class> {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_descend_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the recursive ``descend`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=True)

    return Stripped(
        f"""\
/**
 * Iterate recursively over the instances referenced from this instance.
 *
 * @returns Iterator over the referenced instances
 */
*descend(): IterableIterator<Class> {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_default_value(
    default: intermediate.Default,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the TypeScript code representing the default value of an argument."""
    code: str

    if isinstance(default, intermediate.DefaultPrimitive):
        if default.value is None:
            code = "null"
        elif isinstance(default.value, bool):
            code = typescript_common.boolean_literal(default.value)
        elif isinstance(default.value, int):
            if not typescript_common.representable_as_number(default.value):
                return None, Error(
                    default.parsed.node,
                    f"The value is not representable as a double-precision "
                    f"floating point number: {default.value}",
                )
            code = typescript_common.numeric_literal(default.value)
        elif isinstance(default.value, float):
            code = typescript_common.numeric_literal(default.value)
        elif isinstance(default.value, str):
            code = typescript_common.string_literal(default.value)
        else:
            assert_never(default.value)
    elif isinstance(default, intermediate.DefaultEnumerationLiteral):
        code = ".".join(
            [
                typescript_naming.enum_name(default.enumeration.name),
                typescript_naming.enum_literal_name(default.literal.name),
            ]
        )
    else:
        assert_never(default)

    return Stripped(code), None


@require(lambda cls: not cls.is_implementation_specific)
@require(lambda cls: not cls.constructor.is_implementation_specific)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_constructor(
    cls: intermediate.ClassUnion,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the constructor function for the given concrete class ``cls``.

    Return empty string if there is an empty constructor.
    """
    if (
        len(cls.constructor.arguments) == 0
        and len(cls.constructor.inlined_statements) == 0
    ):
        return Stripped(""), None

    blocks = []  # type: List[str]

    arg_codes = []  # type: List[str]
    for arg in cls.constructor.arguments:
        arg_type = typescript_common.generate_type(type_annotation=arg.type_annotation)
        arg_name = typescript_naming.argument_name(arg.name)

        if arg.default is None:
            arg_codes.append(Stripped(f"{arg_name}: {arg_type}"))
        else:
            default_value, default_value_error = _generate_default_value(arg.default)
            if default_value_error is not None:
                return None, Error(
                    cls.parsed.node,
                    f"Failed to generate the default value for "
                    f"the constructor argument {arg.name!r} of class {cls.name!r}",
                    [default_value_error],
                )

            assert default_value is not None

            arg_codes.append(Stripped(f"{arg_name}: {arg_type} = {default_value}"))

    if len(arg_codes) == 0:
        blocks.append("constructor() {")
    if len(arg_codes) == 1:
        blocks.append(f"constructor({arg_codes[0]}) {{")
    else:
        arg_block = ",\n".join(arg_codes)
        blocks.append(
            Stripped(
                f"""\
constructor(
{I}{indent_but_first_line(arg_block, I)}
) {{"""
            )
        )

    body = ["super();"]  # type: List[str]

    for stmt in cls.constructor.inlined_statements:
        if isinstance(stmt, intermediate_construction.AssignArgument):
            if stmt.default is None:
                body.append(
                    f"this.{typescript_naming.property_name(stmt.name)} = "
                    f"{typescript_naming.argument_name(stmt.argument)};"
                )
            else:
                if isinstance(stmt.default, intermediate_construction.EmptyList):
                    prop = cls.properties_by_name[stmt.name]

                    type_anno = prop.type_annotation
                    while isinstance(type_anno, intermediate.OptionalTypeAnnotation):
                        type_anno = type_anno.value

                    prop_type = typescript_common.generate_type(
                        type_annotation=type_anno
                    )

                    arg_name = typescript_naming.argument_name(stmt.argument)

                    body.append(
                        f"""\
this.{typescript_naming.property_name(stmt.name)} = ({arg_name} !== null)
{I}? {arg_name}
{I}: new {prop_type}();"""
                    )
                elif isinstance(
                    stmt.default, intermediate_construction.DefaultEnumLiteral
                ):
                    literal_code = ".".join(
                        [
                            typescript_naming.enum_name(stmt.default.enum.name),
                            typescript_naming.enum_literal_name(
                                stmt.default.literal.name
                            ),
                        ]
                    )

                    arg_name = typescript_naming.argument_name(stmt.argument)

                    body.append(
                        Stripped(
                            f"""\
this.{typescript_naming.property_name(stmt.name)} = ({arg_name})
{I}? {arg_name}
{I}: {literal_code};"""
                        )
                    )
                else:
                    assert_never(stmt.default)

        else:
            assert_never(stmt)

    blocks.append("\n".join(textwrap.indent(stmt_code, I) for stmt_code in body))

    blocks.append("}")

    return Stripped("\n".join(blocks)), None


@require(lambda cls, prop: id(prop) in cls.property_id_set)
@require(lambda prop: prop.description is not None)
def _generate_comment_for_property(
    cls: intermediate.ClassUnion,
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    # NOTE (mristin, 2022-10-29):
    # We need to write a double assertion for mypy.
    assert prop.description is not None

    # fmt: off
    comment, errors = (
        typescript_description
        .generate_documentation_comment_for_summary_remarks_constraints(
            description=prop.description,
            context=typescript_description.Context(
                module=typescript_common.TYPES_MODULE,
                cls_or_enum=cls,
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_interface(
    interface: intermediate.Interface,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate TypeScript code for the given ``interface``."""
    # Code blocks separated by double newlines and indented once
    blocks = []  # type: List[Stripped]

    # region Properties

    for prop in interface.properties:
        prop_type = typescript_common.generate_type(
            type_annotation=prop.type_annotation
        )
        prop_name = typescript_naming.property_name(prop.name)

        if prop.description is not None:
            (
                prop_comment,
                prop_comment_errors,
            ) = _generate_comment_for_property(cls=interface.base, prop=prop)

            if prop_comment_errors is not None:
                return None, Error(
                    prop.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for the property {prop.name!r}",
                    prop_comment_errors,
                )

            blocks.append(Stripped(f"{prop_comment}\n" f"{prop_name}: {prop_type};"))
        else:
            blocks.append(Stripped(f"{prop_name}: {prop_type};"))

    # endregion

    # region Signatures

    for signature in interface.signatures:
        signature_blocks = []  # type: List[Stripped]

        if signature.description is not None:
            (
                signature_comment,
                signature_comment_errors,
            ) = typescript_description.generate_documentation_comment_for_signature(
                signature.description,
                context=typescript_description.Context(
                    module=typescript_common.TYPES_MODULE, cls_or_enum=interface.base
                ),
            )

            if signature_comment_errors is not None:
                return None, Error(
                    signature.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for signature {signature.name!r}",
                    signature_comment_errors,
                )

            assert signature_comment is not None

            signature_blocks.append(signature_comment)

        # fmt: off
        returns = (
            typescript_common.generate_type(type_annotation=signature.returns)
            if signature.returns is not None else "void"
        )
        # fmt: on

        arg_codes = []  # type: List[Stripped]
        for arg in signature.arguments:
            arg_type = typescript_common.generate_type(
                type_annotation=arg.type_annotation
            )
            arg_name = typescript_naming.argument_name(arg.name)
            arg_codes.append(Stripped(f"{arg_name}: {arg_type}"))

        signature_name = typescript_naming.method_name(signature.name)
        if len(arg_codes) > 2:
            arg_block = ",\n".join(arg_codes)
            signature_blocks.append(
                Stripped(
                    f"""\
{signature_name}(
{I}{indent_but_first_line(arg_block, I)}
): {returns};"""
                )
            )
        elif len(arg_codes) == 1:
            signature_blocks.append(
                Stripped(f"{signature_name}({arg_codes[0]}): {returns};")
            )
        else:
            assert len(arg_codes) == 0
            signature_blocks.append(Stripped(f"{signature_name}(): {returns};"))

        blocks.append(Stripped("\n".join(signature_blocks)))

    # region over_X_or_empty getter

    for prop in interface.properties:
        if isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        ) and isinstance(prop.type_annotation.value, intermediate.ListTypeAnnotation):
            prop_name = typescript_naming.property_name(prop.name)
            items_type = typescript_common.generate_type(
                prop.type_annotation.value.items
            )

            over_x_or_empty = typescript_naming.method_name(
                Identifier(f"over_{prop.name}_or_empty")
            )

            blocks.append(
                Stripped(
                    f"""\
/**
 * Yield from {{@link {prop_name}}} if it is set, or yield nothing.
 */
{over_x_or_empty}(): IterableIterator<{items_type}>;"""
                )
            )

    # endregion

    comment = None  # type: Optional[Stripped]
    if interface.description is not None:
        # fmt: off
        comment, comment_errors = (
            typescript_description
            .generate_documentation_comment_for_summary_remarks_constraints(
                interface.description,
                context=typescript_description.Context(
                    module=typescript_common.TYPES_MODULE,
                    cls_or_enum=interface.base
                )
            )
        )
        # fmt: on

        if comment_errors is not None:
            return None, Error(
                interface.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

    empty_interface = False
    if len(blocks) == 0:
        blocks = [Stripped("// Intentionally empty.")]
        empty_interface = True

    writer = io.StringIO()
    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    name = typescript_naming.interface_name(interface.name)

    if empty_interface:
        writer.write(
            "// eslint-disable-next-line @typescript-eslint/no-empty-interface\n"
        )

    if len(interface.inheritances) == 0:
        writer.write(
            f"""\
export interface {name} extends Class {{
"""
        )
    elif len(interface.inheritances) == 1:
        # NOTE (mristin, 2022-12-07):
        # We can omit Class in the list of inheritances as one of the parents already
        # extends it.
        writer.write(
            f"""\
export interface {name}
{I}extends {typescript_naming.interface_name(interface.inheritances[0].name)} {{
"""
        )
    else:
        # NOTE (mristin, 2022-12-07):
        # We can omit Class in the list of inheritances as one of the parents already
        # extends it.
        writer.write(
            f"""\
export interface {name}
"""
        )

        for i, inheritance in enumerate(interface.inheritances):
            if i == 0:
                writer.write(
                    f"""\
{I}extends {typescript_naming.interface_name(inheritance.name)},
"""
                )
            elif i < len(interface.inheritances) - 1:
                writer.write(
                    f"""\
{II}{typescript_naming.interface_name(inheritance.name)},
"""
                )
            else:
                writer.write(
                    f"""\
{II}{typescript_naming.interface_name(inheritance.name)} {{
"""
                )

    for i, code in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(code, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


@require(lambda cls: not cls.is_implementation_specific)
@require(lambda concrete_cls_index: concrete_cls_index >= 0)
@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_class(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
    concrete_cls_index: int,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate TypeScript code for the given concrete class ``cls``.

    ``concrete_cls_index`` refers to the number of the concrete class in
    the ``concrete_classes`` of the symbol table.
    """
    # NOTE (mristin):
    # Code blocks of the class body separated by double newlines and indented once.

    model_type_enum = typescript_naming.enum_name(Identifier("Model_type"))
    model_type_getter = typescript_naming.method_name(Identifier("model_type"))

    model_type_literal = typescript_naming.enum_literal_name(Identifier(cls.name))

    blocks = [
        Stripped(
            f"""\
/**
 * Indicate the runtime model type of the instance.
 */
{model_type_getter}(): {model_type_enum} {{
{I}// NOTE (mristin, 2022-12-03):
{I}// We yield numbers instead of literals to avoid name lookups on platforms
{I}// which do not provide JIT compilation of hot paths.
{I}return <{model_type_enum}>{concrete_cls_index};  // {model_type_literal}
}}"""
        )
    ]  # type: List[Stripped]

    # region Property definitions

    for prop in cls.properties:
        prop_comment = None  # type: Optional[Stripped]
        if prop.description is not None:
            prop_comment, prop_comment_errors = _generate_comment_for_property(
                cls=cls, prop=prop
            )
            if prop_comment_errors is not None:
                return None, Error(
                    prop.description.parsed.node,
                    "Failed to generate the property comment",
                    prop_comment_errors,
                )
            else:
                assert prop_comment is not None

        prop_type = typescript_common.generate_type(
            type_annotation=prop.type_annotation
        )
        prop_name = typescript_naming.property_name(prop.name)

        writer = io.StringIO()
        if prop_comment is not None:
            writer.write(prop_comment)
            writer.write("\n")
        writer.write(f"{prop_name}: {prop_type};")
        blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region over_X_or_empty getter

    for prop in cls.properties:
        if isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        ) and isinstance(prop.type_annotation.value, intermediate.ListTypeAnnotation):
            prop_name = typescript_naming.property_name(prop.name)
            items_type = typescript_common.generate_type(
                prop.type_annotation.value.items
            )

            over_x_or_empty = typescript_naming.method_name(
                Identifier(f"over_{prop.name}_or_empty")
            )

            blocks.append(
                Stripped(
                    f"""\
/**
 * Yield from {{@link {prop_name}}} if it is set, or yield nothing.
 */
*{over_x_or_empty}(): IterableIterator<{items_type}> {{
{I}if (this.{prop_name} !== null) {{
{II}yield * this.{prop_name};
{I}}}
{I}return;
}}"""
                )
            )

    # endregion

    # region Methods

    errors = []  # type: List[Error]

    for method in cls.methods:
        if isinstance(method, intermediate.ImplementationSpecificMethod):
            implementation_key = specific_implementations.ImplementationKey(
                f"Types/{method.specified_for.name}/{method.name}.ts"
            )

            implementation = spec_impls.get(implementation_key, None)

            if implementation is None:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"The implementation is missing for "
                        f"the implementation-specific method: {implementation_key}",
                    )
                )
                continue

            blocks.append(implementation)
        else:
            errors.append(
                Error(
                    cls.parsed.node,
                    "(mristin, 2022-11-10) "
                    "At the moment, we do not transpile the method body and "
                    "its contracts. We want to finish the meta-model for the V3, "
                    "fix de/serialization and generate SDKs for a couple of languages "
                    "before taking on this rather hard task.",
                )
            )

    blocks.append(_generate_descend_once_method(cls=cls))

    blocks.append(_generate_descend_method(cls=cls))

    visit_name = typescript_naming.method_name(Identifier(f"visit_{cls.name}"))

    blocks.append(
        Stripped(
            f"""\
/**
 * Dispatch `visitor` on this instance.
 *
 * @param visitor - to visit this instance
 */
accept(visitor: AbstractVisitor): void {{
{I}visitor.{visit_name}(this);
}}"""
        )
    )

    visit_with_context_name = typescript_naming.method_name(
        Identifier(f"visit_{cls.name}_with_context")
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Dispatch `visitor` with `context` on this instance.
 *
 * @param visitor - to visit this instance
 * @param context - to be passed along to the dispatched visitor method
 * @typeParam ContextT - type of the context
 */
acceptWithContext<ContextT>(
{I}visitor: AbstractVisitorWithContext<ContextT>,
{I}context: ContextT
) {{
{I}visitor.{visit_with_context_name}(this, context);
}}"""
        )
    )

    transform_name = typescript_naming.method_name(Identifier(f"transform_{cls.name}"))

    blocks.append(
        Stripped(
            f"""\
/**
 * Dispatch the `transformer` on this instance.
 *
 * @param transformer - to transform this instance
 * @returns transformation of this instance
 * @paramType T - type of the transformation result
 */
transform<T>(transformer: AbstractTransformer<T>): T {{
{I}return transformer.{transform_name}(this);
}}"""
        )
    )

    transform_with_context_name = typescript_naming.method_name(
        Identifier(f"transform_{cls.name}_with_context")
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Dispatch the `transformer` on this instance in `context`.
 *
 * @param transformer - to transform this instance
 * @param context - to be passed along to the `transformer`
 * @returns transformation of this instance
 * @paramType T - type of the transformation result
 * @paramType ContextT - type of the transformation context
 */
transformWithContext<ContextT, T>(
{I}transformer: AbstractTransformerWithContext<ContextT, T>,
{I}context: ContextT
): T {{
{I}return transformer.{transform_with_context_name}(
{II}this, context
{I});
}}"""
        )
    )

    # endregion

    # region Constructor

    if cls.constructor.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"Types/{cls.name}/{cls.name}.py"
        )
        implementation = spec_impls.get(implementation_key, None)

        if implementation is None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"The implementation of the implementation-specific constructor "
                    f"is missing: {implementation_key}",
                )
            )
        else:
            blocks.append(implementation)
    else:
        constructor_block, error = _generate_constructor(cls=cls)

        if error is not None:
            errors.append(error)
        else:
            assert constructor_block is not None

            # NOTE (mristin, 2022-06-21):
            # Empty constructor will be automatically generated by the interpreter.
            if constructor_block != "":
                blocks.append(constructor_block)

    # endregion

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the code for the class {cls.name}",
            errors,
        )

    if len(blocks) == 0:
        blocks = [Stripped("// Intentionally empty.")]

    # region Description

    comment = None  # type: Optional[Stripped]
    if cls.description is not None:
        comment, comment_errors = _generate_comment_for_cls_or_enum(cls_or_enum=cls)
        if comment_errors is not None:
            return None, Error(
                cls.description.parsed.node,
                f"Failed to generate the documentation comment for class {cls.name!r}",
                comment_errors,
            )

        assert comment is not None

    # endregion

    # region Implements

    # NOTE (mristin, 2022-12-14):
    # Since JavaScript and hence TypeScript do not support multiple inheritance, we
    # model all the abstract classes as interfaces. Our class therefore extends only
    # the most general class ``Class``, but implements all the corresponding interfaces.
    #
    # Interfaces are only descriptive in TypeScript, so we can not use them directly
    # for efficient type switching at runtime. To implement type switches, we rely
    # on transformer pattern, where we generate the corresponding type-switching
    # transformers. See ``AS_*_TRANSFORMER``'s.

    interface_names = []  # type: List[Identifier]

    if len(cls.concrete_descendants) > 0:
        # NOTE (mristin, 2022-12-14):
        # We do not have to add any other interfaces, as the interface corresponding
        # to this concrete class will already entail all the antecedents.

        assert cls.interface is not None, (
            f"Expected interface for the class {cls.name!r} "
            f"as it has concrete descendants"
        )

        interface_names.append(typescript_naming.interface_name(cls.name))
    else:
        for inheritance in cls.inheritances:
            assert inheritance.interface is not None, (
                f"Expected interface in the parent class {inheritance.name!r} "
                f"of class {cls.name!r}"
            )

            interface_names.append(
                typescript_naming.interface_name(inheritance.interface.name)
            )

    # endregion

    writer = io.StringIO()
    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    name = typescript_naming.class_name(cls.name)

    if len(interface_names) == 0:
        writer.write(
            f"""\
export class {name} extends Class {{
"""
        )
    else:
        interface_names_joined = ",\n".join(interface_names)
        writer.write(
            f"""\
export class {name}
{I}extends Class
{I}implements {indent_but_first_line(interface_names_joined, I)} {{
"""
        )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_abstract_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the abstract visitor."""
    blocks = [
        Stripped(
            f"""\
/**
 * Double-dispatch on `that`.
 */
visit(that: Class): void {{
{I}that.accept(this);
}}"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_name = typescript_naming.method_name(Identifier(f"visit_{cls.name}"))
        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Visit `that`.
 *
 * @param that - instance to be visited
 */
abstract {visit_name}(
{I}that: {cls_name}
): void;"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Visit the instances of the model.
 */
export abstract class AbstractVisitor {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_visitor_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the abstract visitor with context."""
    blocks = [
        Stripped(
            f"""\
/**
 * Double-dispatch on `that` in `context`.
 *
 * @param that - instance to be visited
 * @param context - of the visitation
 */
visitWithContext(
{I}that: Class,
{I}context: ContextT
): void {{
{I}that.acceptWithContext(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_with_context_name = typescript_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )
        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Visit `that` in `context`.
 *
 * @param that - instance to be visited
 * @param context - of the visitation
 */
abstract {visit_with_context_name}(
{I}that: {cls_name},
{I}context: ContextT
): void;"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Visit the instances of the model with context.
 *
 * @typeParam ContextT - type of the visitation context
 */
export abstract class AbstractVisitorWithContext<ContextT> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_pass_through_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the pass-through visitor."""
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_name = typescript_naming.method_name(Identifier(f"visit_{cls.name}"))
        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Visit `that`.
 *
 * @param that - instance to be visited
 */
{visit_name}(
{I}that: {cls_name}
): void {{
{I}for (const another of that.descendOnce()) {{
{II}this.visit(another);
{I}}}
}}"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Visit the instances of the model without action.
 *
 * @remarks
 * This visitor is not meant to be directly used. Instead, you usually
 * inherit from it, and implement only the relevant visit methods.
 */
export class PassThroughVisitor extends AbstractVisitor {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_pass_through_visitor_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the pass-through visitor with context."""
    blocks = [
        Stripped(
            f"""\
/**
 * Double-dispatch on `that` in `context`.
 */
visitWithContext(
{I}that: Class,
{I}context: ContextT
): void {{
{I}that.acceptWithContext(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_with_context_name = typescript_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )
        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Visit `that` in `context`.
 *
 * @param that - instance to be visited
 * @param context - of the visitation
 */
{visit_with_context_name}(
{I}that: {cls_name},
{I}context: ContextT
): void {{
{I}for (const another of that.descendOnce()) {{
{II}this.visitWithContext(another, context);
{I}}}
}}"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Visit the instances of the model without action and in context.
 *
 * @remarks
 * This visitor is not meant to be directly used. Instead, you usually
 * inherit from it, and implement only the relevant visit methods.
 */
export class PassThroughVisitorWithContext<ContextT>
{II}extends AbstractVisitorWithContext<ContextT> {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the abstract transformer."""
    blocks = [
        Stripped(
            f"""\
/**
 * Double-dispatch on `that`.
 */
transform(that: Class): T {{
{I}return that.transform(this);
}}"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}")
        )

        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Transform `that`.
 *
 * @param that - instance to be transformed
 * @returns transformed `that`
 */
abstract {transform_name}(
{I}that: {cls_name}
): T;"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Transform the instance of the model.
 *
 * @typeParam T - type of the transformation result
 */
export abstract class AbstractTransformer<T> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the abstract transformer with context."""
    blocks = [
        Stripped(
            f"""\
/**
 * Double-dispatch on `that` in `context`.
 *
 * @param that - instance to be transformed
 * @param context - of the transformation
 * @returns transformed `that`
 */
transformWithContext(
{I}that: Class,
{I}context: ContextT
): T {{
{I}return that.transformWithContext(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_with_context_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Transform `that` in `context`.
 *
 * @param that - instance to be transformed
 * @param context - of the transformation
 * @returns transformed `that`
 */
abstract {transform_with_context_name}(
{I}that: {cls_name},
{I}context: ContextT
): T;"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Transform the instances of the model in context.
 *
 * @typeParam ContextT - type of the transformation context
 * @typeParam T - type of the transformation result
 */
export abstract class AbstractTransformerWithContext<ContextT, T> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_transformer_with_default(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the transformer with default transformation."""
    blocks = [
        Stripped(
            """\
/**
 * Default value which is returned if no override of the transformation
 */
defaultResult: T"""
        ),
        Stripped(
            f"""\
/**
 * Initialize with the given `default` value.
 *
 * @param defaultResult - returned if no override of the transformation
 */
constructor(defaultResult: T) {{
{I}super();
{I}this.defaultResult = defaultResult;
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}")
        )

        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Transform `that`.
 *
 * @param that - instance to be transformed
 * @returns transformed `that`
 */
/* eslint-disable @typescript-eslint/no-unused-vars */
{transform_name}(
{I}that: {cls_name}
): T {{
{I}return this.defaultResult;
}}
/* eslint-enable @typescript-eslint/no-unused-vars */"""
            )
        )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Transform the instances of the model.
 *
 * @remarks
 * If you do not override the transformation methods, they simply
 * return {@link defaultResult}.
 *
 * @typeParam T - type of the transformation result
 */
export class TransformerWithDefault<T> extends AbstractTransformer<T> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_transformer_with_default_and_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the transformer with default transformation and context."""
    blocks = [
        Stripped(
            """\
/**
 * Default value which is returned if no override of the transformation
 */
defaultResult: T"""
        ),
        Stripped(
            f"""\
/**
 * Initialize with the given `default` value.
 *
 * @param defaultResult - returned if no override of the transformation
 */
constructor(defaultResult: T) {{
{I}super();
{I}this.defaultResult = defaultResult;
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_with_context_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        cls_name = typescript_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
/**
 * Transform `that` in `context`.
 *
 * @param that - instance to be transformed
 * @param context - of the visitation
 * @returns transformed `that`
 */
/* eslint-disable @typescript-eslint/no-unused-vars */
{transform_with_context_name}(
{I}that: {cls_name},
{I}context: ContextT
): T {{
{I}return this.defaultResult;
}}
/* eslint-enable @typescript-eslint/no-unused-vars */"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Transform the instances of the model in context.
 *
 * @remarks
 * If you do not override the transformation methods, they simply
 * return {{@link defaultResult}}.
 *
 * @typeParam ContextT - type of the visitation context
 * @typeParam T - type of the transformation result
 */
export class TransformerWithDefaultAndContext<ContextT, T>
{II}extends AbstractTransformerWithContext<ContextT, T> {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_comment_for_meta_model(
    description: intermediate.DescriptionOfMetaModel,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for the given meta-model."""
    # fmt: off
    comment, errors = (
        typescript_description
        .generate_documentation_comment_for_summary_remarks_constraints(
            description=description,
            context=typescript_description.Context(
                module=typescript_common.TYPES_MODULE, cls_or_enum=None
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


def _generate_as_interface_transformer(
    interface: intermediate.Interface, symbol_table: intermediate.SymbolTable
) -> Stripped:
    blocks = []  # type: List[Stripped]

    interface_name = typescript_naming.interface_name(interface.name)

    for cls in symbol_table.concrete_classes:
        transform_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}")
        )

        cls_name = typescript_naming.class_name(cls.name)

        if cls.is_subclass_of(interface.base):
            blocks.append(
                Stripped(
                    f"""\
{transform_name}(
{I}that: {cls_name}
): {interface_name} | null {{
{I}return that as {interface_name};
}}"""
                )
            )
        else:
            blocks.append(
                Stripped(
                    f"""\
/* eslint-disable @typescript-eslint/no-unused-vars */
{transform_name}(
{I}that: {cls_name}
): {interface_name} | null {{
{I}return null;
}}
/* eslint-enable @typescript-eslint/no-unused-vars */"""
                )
            )

    transformer_name = typescript_naming.class_name(
        Identifier(f"As_{interface.name}_transformer")
    )

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Try to cast an instance of the model to {{@link {interface_name}}}.
 */
class {transformer_name}
{II}extends AbstractTransformer<{interface_name} | null> {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_type_matcher(symbol_table: intermediate.SymbolTable) -> Stripped:
    blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        transform_name = typescript_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        cls_name = typescript_naming.class_name(cls.name)

        is_name = typescript_naming.function_name(Identifier(f"is_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
/* eslint-disable @typescript-eslint/no-unused-vars */
{transform_name}(
{I}that: {cls_name},
{I}other: Class
): boolean {{
{I}return {is_name}(other);
}}
/* eslint-enable @typescript-eslint/no-unused-vars */"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class TypeMatcher extends AbstractTransformerWithContext<
{I}Readonly<Class>,
{I}boolean
> {{"""
    )

    for i, block in enumerate(blocks):
        if i == 0:
            writer.write("\n")
        else:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: VerifiedIntermediateSymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the TypeScript code of the structures based on the symbol table.

    The ``aas_module`` indicates the fully-qualified name of the base module.
    """
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    if symbol_table.meta_model.description is not None:
        # fmt: off
        comment, comment_errors = (
            _generate_comment_for_meta_model(
                description=symbol_table.meta_model.description
            )
        )
        # fmt: on

        if comment_errors is not None:
            errors.extend(comment_errors)
        else:
            assert comment is not None
            blocks.append(comment)

    model_type_getter = typescript_naming.method_name(Identifier("model_type"))

    blocks.extend(
        [
            typescript_common.WARNING,
            _generate_model_type_enum(symbol_table=symbol_table),
            _generate_over_model_type_enum(symbol_table=symbol_table),
            Stripped(
                f"""\
/**
 * Represent the most general class of an AAS model.
 */
export abstract class Class {{
{I}/**
{I} * Indicate the runtime model type of an instance.
{I} */
{I}abstract {model_type_getter}(): ModelType;

{I}/**
{I} * Iterate over all the instances referenced from this one.
{I} */
{I}abstract descendOnce(): IterableIterator<Class>;

{I}/**
{I} * Iterate recursively over all the instances referenced from this one.
{I} */
{I}abstract descend(): IterableIterator<Class>;

{I}/**
{I} * Dispatch the `visitor` on this instance.
{I} *
{I} * @param visitor - to be dispatched
{I} */
{I}abstract accept(visitor: AbstractVisitor): void;

{I}/**
{I} * Dispatch the `visitor` on this instance with `context`.
{I} *
{I} * @param visitor - to be dispatched
{I} * @param context - of the visitation
{I} * @typeParam ContextT - type of the visitation context
{I} */
{I}abstract acceptWithContext<ContextT>(
{II}visitor: AbstractVisitorWithContext<ContextT>,
{II}context: ContextT
{I}): void;

{I}/**
{I} * Dispatch the `transformer` on this instance.
{I} *
{I} * @param transformer - to be dispatched
{I} * @return this instance transformed
{I} * @typeParam T - type of the transformation result
{I} */
{I}abstract transform<T>(transformer: AbstractTransformer<T>): T;

{I}/**
{I} * Dispatch the `transformer` on this instance in `context`.
{I} *
{I} * @param transformer - to be dispatched
{I} * @param context - of the transformation
{I} * @return this instance transformed
{I} * @typeParam T - type of the transformation result
{I} */
{I}abstract transformWithContext<ContextT, T>(
{II}transformer: AbstractTransformerWithContext<ContextT, T>,
{II}context: ContextT
{I}): T;
}}"""
            ),
        ]
    )

    concrete_class_to_index = {
        concrete_cls: i for i, concrete_cls in enumerate(symbol_table.concrete_classes)
    }

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            block, error = _generate_enum(enum=our_type)
            if error is not None:
                errors.append(error)
                continue
            else:
                assert block is not None
                blocks.append(block)
                blocks.append(_generate_over_enum(enum=our_type))

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-11-10):
            # We do not generate the constrained primitives as types. We only
            # consider them in the verification.
            continue

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Types/{our_type.name}.ts"
                )

                block = spec_impls.get(implementation_key, None)
                if block is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The implementation is missing "
                            f"for the implementation-specific "
                            f"class: {implementation_key}",
                        )
                    )
                else:
                    blocks.append(block)
            else:
                if our_type.interface is not None:
                    block, error = _generate_interface(interface=our_type.interface)
                    if error is not None:
                        errors.append(error)
                    else:
                        assert block is not None
                        blocks.append(block)

                if isinstance(our_type, intermediate.ConcreteClass):
                    block, error = _generate_class(
                        cls=our_type,
                        spec_impls=spec_impls,
                        concrete_cls_index=concrete_class_to_index[our_type],
                    )
                    if error is not None:
                        errors.append(error)
                    else:
                        assert block is not None
                        blocks.append(block)
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            _generate_abstract_visitor(symbol_table=symbol_table),
            _generate_abstract_visitor_with_context(symbol_table=symbol_table),
            _generate_pass_through_visitor(symbol_table=symbol_table),
            _generate_pass_through_visitor_with_context(symbol_table=symbol_table),
            _generate_abstract_transformer(symbol_table=symbol_table),
            _generate_abstract_transformer_with_context(symbol_table=symbol_table),
            _generate_transformer_with_default(symbol_table=symbol_table),
            _generate_transformer_with_default_and_context(symbol_table=symbol_table),
        ]
    )

    for cls in symbol_table.classes:
        if cls.interface is not None:
            transformer_name = typescript_naming.class_name(
                Identifier(f"As_{cls.interface.name}_transformer")
            )
            constant_transformer = typescript_naming.constant_name(
                Identifier(f"As_{cls.interface.name}_transformer")
            )

            as_interface = typescript_naming.function_name(
                Identifier(f"as_{cls.interface.name}")
            )

            is_interface = typescript_naming.function_name(
                Identifier(f"is_{cls.interface.name}")
            )

            interface_name = typescript_naming.interface_name(cls.interface.name)

            blocks.extend(
                [
                    _generate_as_interface_transformer(
                        interface=cls.interface, symbol_table=symbol_table
                    ),
                    Stripped(
                        f"""\
const {constant_transformer} =
{I}new {transformer_name}();"""
                    ),
                    Stripped(
                        f"""\
/**
 * Try to cast `that` instance to
 * the interface {{@link {interface_name}}}.
 *
 * @param that - instance to be casted
 * @returns - casted `that` if cast successful, or `null`
 */
export function {as_interface}(
{I}that: Class
): {interface_name} | null {{
{I}return {constant_transformer}.transform(that);
}}"""
                    ),
                    Stripped(
                        f"""\
/**
 * Check the type of `that` instance.
 *
 * @param that - instance to be type-checked
 * @returns `true` if the type check is successful
 */
export function {is_interface}(
{I}that: Class
): that is {interface_name} {{
{I}return {as_interface}(that) !== null;
}}"""
                    ),
                ]
            )

        # NOTE (mristin, 2022-11-23):
        # We add these functions to make a uniform interface to ``isX`` checks.
        # Without these functions, the clients would have to refactor a lot once
        # the meta-model changes and a class gets descendants where it had none before.
        if (
            isinstance(cls, intermediate.ConcreteClass)
            and len(cls.concrete_descendants) == 0
        ):
            cls_name = typescript_naming.class_name(cls.name)

            as_cls = typescript_naming.function_name(Identifier(f"as_{cls.name}"))

            is_cls = typescript_naming.function_name(Identifier(f"is_{cls.name}"))

            blocks.extend(
                [
                    Stripped(
                        f"""\
/**
 * Try to cast `that` instance to
 * the class {{@link {cls_name}}}.
 *
 * @param that - instance to be casted
 * @returns - casted `that` if cast successful, or `null`
 */
export function {as_cls}(
{I}that: Class
): {cls_name} | null {{
{I}return (that instanceof {cls_name})
{II}? <{cls_name}>that
{II}: null;
}}"""
                    ),
                    Stripped(
                        f"""\
/**
 * Check the type of `that` instance.
 *
 * @param that - instance to be type-checked
 * @returns `true` if the type check is successful
 */
export function {is_cls}(
{I}that: Class
): that is {cls_name} {{
{I}return that instanceof {cls_name};
}}"""
                    ),
                ]
            )

    blocks.append(_generate_type_matcher(symbol_table=symbol_table))
    blocks.append(
        Stripped(
            """\
const TYPE_MATCHER = new TypeMatcher();"""
        )
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Check whether the type of `that` matches the type of `other` instance.
 *
 * @remarks
 * We check with `is*` function. Hence, if the class of `other` is a subclass of
 * the class of `that`, we confirm the match.
 *
 * @param that - standard instance
 * @param other - instance whose type is compared against `that`
 */
export function typesMatch<ClassT extends Class>(
{I}that: ClassT,
{I}other: Class
): other is ClassT {{
{I}return TYPE_MATCHER.transformWithContext(that, other);
}}"""
        )
    )

    blocks.append(typescript_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Generate TypeScript code for de/serialization of enumerations."""

import io
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, Stripped, Identifier, indent_but_first_line
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)
from aas_core_codegen import naming


def _generate_model_type_from_string(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the function to de-serialize model type from a string."""
    model_type_enum = typescript_naming.enum_name(Identifier("Model_type"))

    keys_values = []  # type: List[Stripped]
    for concrete_cls in symbol_table.concrete_classes:
        json_model_type = naming.json_model_type(concrete_cls.name)
        model_type_literal = typescript_naming.enum_literal_name(concrete_cls.name)

        assert json_model_type == model_type_literal, (
            f"Expected the JSON model type for class {concrete_cls.name!r}, "
            f"{json_model_type!r}, to equal the literal in the enumeration "
            f"{model_type_enum!r}, but it does not. "
            f"The literal is: {model_type_literal!r}. This will make it very confusing "
            f"for the user if the model type in JSON and in TypeScript differ. "
            f"Please contact the developers and re-evaluate whether it makes sense "
            f"to change the naming of the TypeScript enumeration literals."
        )

        json_model_type_literal = typescript_common.string_literal(json_model_type)

        keys_values.append(
            Stripped(
                f"""\
[
{I}{json_model_type_literal},
{I}AasTypes.{model_type_enum}.{model_type_literal}
]"""
            )
        )

    map_name = typescript_naming.constant_name(Identifier("model_type_from_string"))
    keys_values_joined = ",\n".join(keys_values)

    from_string = typescript_naming.function_name(Identifier("model_type_from_string"))

    return [
        Stripped(
            f"""\
const {map_name} = new Map<string, AasTypes.{model_type_enum}>([
{I}{indent_but_first_line(keys_values_joined, I)}
]);"""
        ),
        Stripped(
            f"""\
/**
 * Parse `text` as a string representation of {{@link types!{model_type_enum}}}.
 *
 * @param text - string representation of {{@link types!{model_type_enum}}}
 * @returns literal of {{@link types!{model_type_enum}}}, if valid, and `null` otherwise
 */
export function {from_string}(
{I}text: string
): AasTypes.{model_type_enum} | null {{
{I}const result = {map_name}.get(text);
{I}return result !== undefined ? result : null;
}}"""
        ),
    ]


def _generate_model_type_to_string(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the function to serialize a runtime model type to a string."""
    model_type_enum = typescript_naming.enum_name(Identifier("Model_type"))

    keys_values = []  # type: List[Stripped]
    for concrete_cls in symbol_table.concrete_classes:
        json_model_type = naming.json_model_type(concrete_cls.name)
        model_type_literal = typescript_naming.enum_literal_name(concrete_cls.name)

        assert json_model_type == model_type_literal, (
            f"Expected the JSON model type for class {concrete_cls.name!r}, "
            f"{json_model_type!r}, to equal the literal in the enumeration "
            f"{model_type_enum!r}, but it does not. "
            f"The literal is: {model_type_literal!r}. This will make it very confusing "
            f"for the user if the model type in JSON and in TypeScript differ. "
            f"Please contact the developers and re-evaluate whether it makes sense "
            f"to change the naming of the TypeScript enumeration literals."
        )

        json_model_type_literal = typescript_common.string_literal(json_model_type)

        keys_values.append(
            Stripped(
                f"""\
[
{I}AasTypes.{model_type_enum}.{model_type_literal},
{I}{json_model_type_literal}
]"""
            )
        )

    map_name = typescript_naming.constant_name(Identifier("model_type_to_string"))
    keys_values_joined = ",\n".join(keys_values)

    to_string = typescript_naming.function_name(Identifier("model_type_to_string"))
    must_to_string = typescript_naming.function_name(
        Identifier("must_model_type_to_string")
    )

    return [
        Stripped(
            f"""\
const {map_name} = new Map<AasTypes.{model_type_enum}, string>([
{I}{indent_but_first_line(keys_values_joined, I)}
]);"""
        ),
        Stripped(
            f"""\
/**
 * Translate {{@link types!{model_type_enum}}} to a string.
 *
 * @param value - to be stringified
 * @returns string representation of {{@link types!{model_type_enum}}},
 * if `value` valid, and `null` otherwise
 */
export function {to_string}(
{I}value: AasTypes.{model_type_enum}
): string | null {{
{I}const result = {map_name}.get(value);
{I}return result !== undefined ? result : null;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Translate {{@link types!{model_type_enum}}} to a string.
 *
 * @param value - to be stringified
 * @returns string representation of {{@link types!{model_type_enum}}}
 * @throws
 * {{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error|Error}}
 * if the `value` is invalid
 */
export function {must_to_string}(
{I}value: AasTypes.{model_type_enum}
): string {{
{I}const result = {map_name}.get(value);
{I}if (result === undefined) {{
{II}throw new Error(
{III}`Invalid literal of {model_type_enum}: ${{value}}`
{II});
{I}}}
{I}return result;
}}"""
        ),
    ]


def _generate_enum_from_string(enumeration: intermediate.Enumeration) -> Stripped:
    """Generate the functions for de-serializing enumeration from strings."""
    blocks = []  # type: List[Stripped]

    name = typescript_naming.enum_name(enumeration.name)

    # region From-string-map

    items = []  # type: List[str]
    for literal in enumeration.literals:
        literal_name = typescript_naming.enum_literal_name(literal.name)
        literal_value = typescript_common.string_literal(literal.value)

        items.append(f"[{literal_value}, AasTypes.{name}.{literal_name}]")

    from_str_map_name = typescript_naming.constant_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const {from_str_map_name} = new Map<string, AasTypes.{name}>([
{I}{indent_but_first_line(items_joined, I)}
]);"""
        )
    )

    # endregion

    # region From-string-function

    from_str_name = typescript_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Parse `text` as a string representation of {{@link types!{name}}}.
 *
 * @param text - string representation of {{@link types!{name}}}
 * @returns literal of {{@link types!{name}}}, if valid, and `null` otherwise
 */
export function {from_str_name}(
{I}text: string
): AasTypes.{name} | null {{
{I}const result = {from_str_map_name}.get(text);
{I}return result !== undefined ? result : null;
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


def _generate_enum_to_string(enumeration: intermediate.Enumeration) -> Stripped:
    """Generate the functions for serializing enumerations to strings."""
    blocks = []  # type: List[Stripped]

    name = typescript_naming.enum_name(enumeration.name)

    # region To-string-map

    items = []  # type: List[str]
    for literal in enumeration.literals:
        literal_name = typescript_naming.enum_literal_name(literal.name)
        literal_value = typescript_common.string_literal(literal.value)

        items.append(f"[AasTypes.{name}.{literal_name}, {literal_value}]")

    to_str_map_name = typescript_naming.constant_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const {to_str_map_name} = new Map<AasTypes.{name}, string>([
{I}{indent_but_first_line(items_joined, I)}
]);"""
        )
    )

    # endregion

    # region To-string-function

    to_str_name = typescript_naming.function_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Translate {{@link types!{name}}} to a string.
 *
 * @param value - to be stringified
 * @returns string representation of {{@link types!{name}}}, if `value` valid, and `null` otherwise
 */
export function {to_str_name}(
{I}value: AasTypes.{name}
): string | null {{
{I}const result = {to_str_map_name}.get(value);
{I}return result !== undefined ? result : null;
}}"""
        )
    )

    must_to_str_name = typescript_naming.function_name(
        Identifier(f"must_{enumeration.name}_to_string")
    )

    blocks.append(
        Stripped(
            f"""\
/**
 * Translate {{@link types!{name}}} to a string.
 *
 * @param value - to be stringified
 * @returns string representation of {{@link types!{name}}}
 * @throws
 * {{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error|Error}}
 * if the `value` is invalid
 */
export function {must_to_str_name}(
{I}value: AasTypes.{name}
): string {{
{I}const result = {to_str_map_name}.get(value);
{I}if (result === undefined) {{
{II}throw new Error(
{III}`Invalid literal of {name}: ${{value}}`
{II});
{I}}}
{I}return result;
}}"""
        )
    )

    # endregion

    return Stripped("\n\n".join(blocks))


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the TypeScript code for the de/serialization of strings."""
    blocks = [
        Stripped(
            """\
/**
 * De/serialize enumerations from and to string representations.
 */"""
        ),
        typescript_common.WARNING,
        Stripped('import * as AasTypes from "./types";'),
        *_generate_model_type_from_string(symbol_table=symbol_table),
        *_generate_model_type_to_string(symbol_table=symbol_table),
    ]

    for enum in symbol_table.enumerations:
        blocks.append(_generate_enum_from_string(enumeration=enum))
        blocks.append(_generate_enum_to_string(enumeration=enum))

    blocks.append(typescript_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None
